# Copyright 2023 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_concurrency import lockutils
from oslo_log import log as logging

from ovn_bgp_agent import constants
from ovn_bgp_agent.drivers.openstack import nb_exceptions
from ovn_bgp_agent.drivers.openstack.utils import common as common_utils
from ovn_bgp_agent.drivers.openstack.utils import driver_utils
from ovn_bgp_agent.drivers.openstack.utils import loadbalancer as lb_utils
from ovn_bgp_agent.drivers.openstack.utils import port as port_utils
from ovn_bgp_agent.drivers.openstack.utils import router as router_utils
from ovn_bgp_agent.drivers.openstack.watchers import base_watcher
from ovn_bgp_agent import exceptions


LOG = logging.getLogger(__name__)
_SYNC_STATE_LOCK = lockutils.ReaderWriterLock()


class LogicalSwitchPortProviderCreateEvent(base_watcher.LSPChassisEvent):
    def __init__(self, bgp_agent):
        events = (self.ROW_UPDATE,)
        super(LogicalSwitchPortProviderCreateEvent, self).__init__(
            bgp_agent, events)

    def match_fn(self, event, row, old):
        '''Match port updates to see if we should expose this lsp

        If the event matches the following criteria, we should
        totally ignore this event, since it is not meant for this host.

        1. this host does not own this lsp
        2. the lsp is not up
        3. the logical switch is not exposed with agent, which means it
           is not a provider network

        When the event still has not been rejected, then the only thing to
        do is to check if the ips for this lsp have not been exported yet.
        '''
        if row.type not in [constants.OVN_VM_VIF_PORT_TYPE,
                            constants.OVN_VIRTUAL_VIF_PORT_TYPE]:
            return
        try:
            # single and dual-stack format
            if not port_utils.has_ip_address_defined(row.addresses[0]):
                return False

            current_chassis = self._get_chassis(row)
            logical_switch = common_utils.get_from_external_ids(
                row, constants.OVN_LS_NAME_EXT_ID_KEY)

            if logical_switch in self.agent.ovn_local_lrps:
                # This is a tenant network, routed through lrp, handled by
                # event LogicalSwitchPortTenantCreateEvent
                return False

            # Check for rejection criteria
            if (current_chassis != self.agent.chassis or
                    not bool(row.up[0]) or
                    not self.agent.is_ls_provider(logical_switch)):
                return False

            # At this point, the port is bound on this host, it is up and
            # the logical switch is exposable by the agent.
            # Only create the ips if not already exposed.
            try:
                ips = port_utils.get_ips_from_lsp(row)
            except exceptions.IpAddressNotFound:
                return False

            return not self.agent.is_ip_exposed(logical_switch, ips)

        except (IndexError, AttributeError):
            return False

    def _run(self, event, row, old):
        with _SYNC_STATE_LOCK.read_lock():
            ips = port_utils.get_ips_from_lsp(row)
            ips_info = port_utils.make_lsp_dict(row)
            self.agent.expose_ip(ips, ips_info)


class LogicalSwitchPortProviderDeleteEvent(base_watcher.LSPChassisEvent):
    def __init__(self, bgp_agent):
        events = (self.ROW_UPDATE, self.ROW_DELETE,)
        super(LogicalSwitchPortProviderDeleteEvent, self).__init__(
            bgp_agent, events)

    def match_fn(self, event, row, old):
        '''Match port deletes or port downs or migrations

        1. [DELETE] Port has been deleted, and we're hosting it
        2. [UPDATE] Port went down, withdraw if we announced it
        3. [UPDATE] Port has been migrated away and we're hosting it
        '''
        if row.type not in [constants.OVN_VM_VIF_PORT_TYPE,
                            constants.OVN_VIRTUAL_VIF_PORT_TYPE]:
            return
        try:
            # single and dual-stack format
            if not port_utils.has_ip_address_defined(row.addresses[0]):
                return False

            try:
                ips = port_utils.get_ips_from_lsp(row)
            except exceptions.IpAddressNotFound:
                return False

            logical_switch = common_utils.get_from_external_ids(
                row, constants.OVN_LS_NAME_EXT_ID_KEY)

            if logical_switch in self.agent.ovn_local_lrps:
                # This is a tenant network, routed through lrp, handled by
                # event LogicalSwitchPortTenantDeleteEvent
                return False

            # Do nothing if we do not expose the current port
            if not self.agent.is_ip_exposed(logical_switch, ips):
                return False

            # Delete event, always execute (since we expose it)
            if event == self.ROW_DELETE:
                return True

            current_chassis = self._get_chassis(row)
            # Delete the port from current chassis, if
            # 1. port went down (while only attached here)
            if (hasattr(old, 'up') and bool(old.up[0]) and   # port was up
                    not bool(row.up[0]) and                  # is now down
                    not port_utils.has_additional_binding(row)):  # and bound
                return True

            # 2. port no longer bound here
            return current_chassis != self.agent.chassis

        except (IndexError, AttributeError):
            return False

    def _run(self, event, row, old):
        with _SYNC_STATE_LOCK.read_lock():
            ips = port_utils.get_ips_from_lsp(row)
            ips_info = port_utils.make_lsp_dict(row)
            self.agent.withdraw_ip(ips, ips_info)


class LogicalSwitchPortFIPCreateEvent(base_watcher.LSPChassisEvent):
    '''Floating IP create events based on the LogicalSwitchPort

    The LSP has information about the host is should be exposed to, which
    adds a bit of complexity in the event match, but saves a lot of queries
    to the OVN NB DB.

    Should trigger on:
    - floating ip was attached to a lsp (external_ids.neutron:port_fip
                                         appeared with information)
    - port with floating ip attached was set to up (old.up = false and
                                                    row.up = true)

    During a migration of a lsp, the following events happen (chronologically):
    1. options.requested_chassis is updated (now a comma separated list)
       we also get external_ids, but only revision_number is updated.
    2. update with only external_ids update (with only a revnum update)
    3. port is set down (by ovn-controller on source host)
    4. update with only external_ids update (with only a revnum update)
    5. external_ids update (neutron:host_id is removed)
    6. options.requested_chassis is updated (with only dest host)
       and external_ids update which now includes neutron:host_id again
    7. port is set up (by ovn-controller on dest host)
    8 and 9 are only a revnum update in the external_ids

    So for migration flow we are only interested in event 7.
    Otherwise the floating ip would be added upon event 2, deleted with
        event 3 and re-added with event 7.

    For the live migration of a VIP (floating ip attached to virtual port),
    the following events happen:
    1. port is set down (by ovn-controller on source host)
    2. external_ids update (neutron:host_id is removed)
    3. port is set up (by ovn-controller on dest host)
    4. external_ids update (neutron:host_id is added)

    In this case we only need to catch event 4.
    '''
    def __init__(self, bgp_agent):
        events = (self.ROW_UPDATE,)
        super(LogicalSwitchPortFIPCreateEvent, self).__init__(
            bgp_agent, events)

    def match_fn(self, event, row, old):
        if row.type not in [constants.OVN_VM_VIF_PORT_TYPE,
                            constants.OVN_VIRTUAL_VIF_PORT_TYPE]:
            return
        try:
            # single and dual-stack format
            if not port_utils.has_ip_address_defined(row.addresses[0]):
                return False

            current_chassis = self._get_chassis(row)
            current_port_fip = row.external_ids.get(
                constants.OVN_FIP_EXT_ID_KEY)
            if (current_chassis != self.agent.chassis or
                    not bool(row.up[0]) or not current_port_fip):
                # Port is not bound on this host, is down or does not have a
                # floating ip attached.
                return False

            if hasattr(old, 'up') and not bool(old.up[0]):
                # Port changed up, which happens when the port is picked up
                # on this host by the ovn-controller during migrations
                return True

            old_port_fip = getattr(old, 'external_ids', {}).get(
                constants.OVN_FIP_EXT_ID_KEY)
            if old_port_fip == current_port_fip:
                # Only if the floating ip has changed (for example from empty
                # to something else) we need to process this update.
                # If nothing else changed in the external_ids, we do not care
                # as it would just cause unnecessary events during migrations.
                # Only case we are interested in is if the chassis has changed
                # with this event.
                # (see the docstring of this class)
                old_chassis = self._get_chassis(old)
                return old_chassis != current_chassis

            # Check if the current port_fip has not been exposed yet
            return not self.agent.is_ip_exposed(
                common_utils.get_from_external_ids(
                    row, constants.OVN_LS_NAME_EXT_ID_KEY),
                current_port_fip)

        except (IndexError, AttributeError):
            return False

    def _run(self, event, row, old):
        try:
            external_ip, external_mac, ls_name = (
                self.agent.get_port_external_ip_and_ls(row.name))
        except nb_exceptions.NATNotFound as e:
            LOG.debug("Logical Switch Port %s does not have all data required"
                      " in its NAT entry: %s", row.name, e)
            return

        with _SYNC_STATE_LOCK.read_lock():
            self.agent.expose_fip(external_ip, external_mac, ls_name, row)


class LogicalSwitchPortFIPDeleteEvent(base_watcher.LSPChassisEvent):
    '''Floating IP delete events based on the LogicalSwitchPort

    The LSP has information about the host is should be exposed to, which
    adds a bit of complexity in the event match, but saves a lot of queries
    to the OVN NB DB.

    Should trigger on:
    - lsp deleted and bound on this host
    - floating ip removed from a lsp (external_ids.neutron:port_fip
                                      disappeared with information)
    - port with floating ip attached was set to down (old.up = true and
                                                      row.up = false)
    - current floating ip is not the same as old floating ip
    '''
    def __init__(self, bgp_agent):
        events = (self.ROW_UPDATE, self.ROW_DELETE,)
        super(LogicalSwitchPortFIPDeleteEvent, self).__init__(
            bgp_agent, events)

    def match_fn(self, event, row, old):
        '''Match port deletes or port downs or migrations or fip changes

        1. [DELETE] Port has been deleted, and we're hosting it
        2. [UPDATE] Port went down, withdraw if we announced it
        3. [UPDATE] Floating IP has been disassociated (or re-associated
                    with another floating ip)
        4. [UPDATE] Port has been migrated away and we're hosting it
        '''
        if row.type not in [constants.OVN_VM_VIF_PORT_TYPE,
                            constants.OVN_VIRTUAL_VIF_PORT_TYPE]:
            return
        try:
            # single and dual-stack format
            if not port_utils.has_ip_address_defined(row.addresses[0]):
                return False

            current_port_fip = port_utils.get_fip(row)
            old_port_fip = port_utils.get_fip(old)
            if not current_port_fip and not old_port_fip:
                # This port is not a floating ip update
                return False

            logical_switch = common_utils.get_from_external_ids(
                row, constants.OVN_LS_NAME_EXT_ID_KEY)
            is_exposed = self.agent.is_ip_exposed(logical_switch,
                                                  old_port_fip or
                                                  current_port_fip)
            if not is_exposed:
                # already deleted or not exposed.
                return False

            # From here on we know we are exposing a FIP (either old or
            #                                             current)

            if event == self.ROW_DELETE:
                # Port is deleting
                return True

            if (hasattr(old, 'up') and bool(old.up[0]) and  # port was up
                    not bool(row.up[0])):                   # is now down
                return True

            if old_port_fip is not None and current_port_fip != old_port_fip:
                # fip has changed, we should remove the old one.
                return True

            # If we reach here, just check if host changed
            current_chassis = self._get_chassis(row)
            return current_chassis != self.agent.chassis

        except (IndexError, AttributeError):
            return False

    def _run(self, event, row, old):
        # First check to remove the fip provided in old (since this might
        # have been updated)
        fip = port_utils.get_fip(old)
        if not fip:
            # Remove the fip provided in the current row, probably a
            # disassociate of the fip (or a down or a move)
            fip = port_utils.get_fip(row)
        if not fip:
            return
        with _SYNC_STATE_LOCK.read_lock():
            self.agent.withdraw_fip(fip, row)


class LogicalSwitchUpdateEvent(base_watcher.LogicalSwitchChassisEvent):
    '''Event to trigger on logical switch vrf config updates'''
    def __init__(self, bgp_agent):
        events = (self.ROW_UPDATE, self.ROW_DELETE)
        super(LogicalSwitchUpdateEvent, self).__init__(
            bgp_agent, events)

    def match_fn(self, event, row, old):
        '''Match updates for vrf configuration

        Will trigger whenever external_ids[neutron_bgpvpn:vni] and
        external_ids[neutron_bgpvpn:type] have been set and either one has
        been updated
        '''

        settings = driver_utils.get_port_vrf_settings(row)
        if settings and event == self.ROW_DELETE:
            # Always run sync method if we are deleting this network (and it
            # had settings applied)
            return True

        old_settings = driver_utils.get_port_vrf_settings(old)
        if old_settings is None:
            # it was not provided in old, so do not process this update
            return False

        return settings != old_settings

    def _run(self, event, row, old):
        with _SYNC_STATE_LOCK.read_lock():
            # NOTE(mnederlof): For now it makes sense to run the sync method
            # as this is triggered with a configured interval anyway and it
            # will add/remove the triggered logical switch.
            # It might make sense in the future to optimize this behaviour.
            self.agent.sync()


class LocalnetCreateDeleteEvent(base_watcher.LSPChassisEvent):
    def __init__(self, bgp_agent):
        events = (self.ROW_CREATE, self.ROW_DELETE,)
        super(LocalnetCreateDeleteEvent, self).__init__(
            bgp_agent, events)

    def match_fn(self, event, row, old):
        if row.type == constants.OVN_LOCALNET_VIF_PORT_TYPE:
            return True
        return False

    def _run(self, event, row, old):
        with _SYNC_STATE_LOCK.read_lock():
            self.agent.sync()


class ChassisRedirectCreateEvent(base_watcher.LRPChassisEvent):
    def __init__(self, bgp_agent):
        events = (self.ROW_UPDATE,)
        super(ChassisRedirectCreateEvent, self).__init__(
            bgp_agent, events)

    def match_fn(self, event, row, old):
        try:
            if not row.networks:
                return False

            # check if hosting-chassis is being added
            hosting_chassis = row.status.get(constants.OVN_STATUS_CHASSIS)
            if hosting_chassis != self.agent.chassis_id:
                # No chassis set or different one
                return False

            if hasattr(old, 'status'):
                # status has changed
                old_hosting_chassis = old.status.get(
                    constants.OVN_STATUS_CHASSIS)
                if old_hosting_chassis != hosting_chassis:
                    return True
        except (IndexError, AttributeError):
            return False
        return False

    def _run(self, event, row, old):
        with _SYNC_STATE_LOCK.read_lock():
            ips_info = port_utils.make_lrp_dict(row)
            ips = [net.split("/")[0] for net in row.networks]
            self.agent.expose_ip(ips, ips_info)


class ChassisRedirectDeleteEvent(base_watcher.LRPChassisEvent):
    def __init__(self, bgp_agent):
        events = (self.ROW_UPDATE, self.ROW_DELETE,)
        super(ChassisRedirectDeleteEvent, self).__init__(
            bgp_agent, events)

    def match_fn(self, event, row, old):
        try:
            if not row.networks:
                return

            if event == self.ROW_DELETE:
                return (row.status.get(constants.OVN_STATUS_CHASSIS) ==
                        self.agent.chassis_id)
            # ROW UPDATE EVENT
            if hasattr(old, 'status'):
                # status has changed
                hosting_chassis = row.status.get(constants.OVN_STATUS_CHASSIS)
                old_hosting_chassis = old.status.get(
                    constants.OVN_STATUS_CHASSIS)
                if (hosting_chassis != old_hosting_chassis and
                        old_hosting_chassis == self.agent.chassis_id):
                    return True
        except (IndexError, AttributeError):
            return False
        return False

    def _run(self, event, row, old):
        with _SYNC_STATE_LOCK.read_lock():
            ips_info = port_utils.make_lrp_dict(row)
            ips = [net.split("/")[0] for net in row.networks]
            self.agent.withdraw_ip(ips, ips_info)


class LogicalSwitchPortSubnetAttachEvent(base_watcher.LSPChassisEvent):
    def __init__(self, bgp_agent):
        events = (self.ROW_UPDATE,)
        super(LogicalSwitchPortSubnetAttachEvent, self).__init__(
            bgp_agent, events)

    def match_fn(self, event, row, old):
        try:
            if row.type != constants.OVN_ROUTER_PORT_TYPE:
                return False
            # skip route_gateway port events
            row_device_owner = row.external_ids.get(
                constants.OVN_DEVICE_OWNER_EXT_ID_KEY)
            if row_device_owner != constants.OVN_ROUTER_INTERFACE:
                return False

            if not bool(row.up[0]):
                return False

            associated_router = row.external_ids.get(
                constants.OVN_DEVICE_ID_EXT_ID_KEY)

            if associated_router not in self.agent.ovn_local_cr_lrps:
                return False

            if hasattr(old, 'up') and not bool(old.up[0]):
                return True

            if hasattr(old, 'external_ids'):
                previous_associated_router = old.external_ids.get(
                    constants.OVN_DEVICE_ID_EXT_ID_KEY)
                if (associated_router != previous_associated_router and
                        previous_associated_router not in
                        self.agent.ovn_local_cr_lrps):
                    return True
        except (IndexError, AttributeError):
            return False
        return False

    def _run(self, event, row, old):
        with _SYNC_STATE_LOCK.read_lock():
            ips = row.external_ids.get(constants.OVN_CIDRS_EXT_ID_KEY,
                                       "").split()
            subnet_info = {
                'associated_router': row.external_ids.get(
                    constants.OVN_DEVICE_ID_EXT_ID_KEY),
                'network': common_utils.get_from_external_ids(
                    row, constants.OVN_LS_NAME_EXT_ID_KEY),
                'address_scopes': driver_utils.get_addr_scopes(row)}
            self.agent.expose_subnet(ips, subnet_info)


class LogicalSwitchPortSubnetDetachEvent(base_watcher.LSPChassisEvent):
    def __init__(self, bgp_agent):
        events = (self.ROW_UPDATE, self.ROW_DELETE,)
        super(LogicalSwitchPortSubnetDetachEvent, self).__init__(
            bgp_agent, events)

    def match_fn(self, event, row, old):
        try:
            if row.type != constants.OVN_ROUTER_PORT_TYPE:
                return False
            # skip route_gateway port events
            row_device_owner = row.external_ids.get(
                constants.OVN_DEVICE_OWNER_EXT_ID_KEY)
            if row_device_owner != constants.OVN_ROUTER_INTERFACE:
                return False

            associated_router = row.external_ids.get(
                constants.OVN_DEVICE_ID_EXT_ID_KEY)

            if event == self.ROW_DELETE:
                if not bool(row.up[0]):
                    return False
                if associated_router in self.agent.ovn_local_cr_lrps:
                    return True
                return False

            # ROW UPDATE
            # We need to withdraw the subnet in the next cases:
            # 1. same/local associated router and status moves from up to down
            # 2. status changes to down and also associated router changes to a
            #    non local one
            # 3. status is up (same) but associated router changes to a non
            #    local one
            if hasattr(old, 'up'):
                if not bool(old.up[0]):
                    return False
                if hasattr(old, 'external_ids'):
                    previous_associated_router = old.external_ids.get(
                        constants.OVN_DEVICE_ID_EXT_ID_KEY)
                    if previous_associated_router in (
                            self.agent.ovn_local_cr_lrps):
                        return True
                else:
                    if associated_router in self.agent.ovn_local_cr_lrps:
                        return True
            else:
                # no change in status
                if not bool(row.up[0]):
                    # it was not exposed
                    return False
                if hasattr(old, 'external_ids'):
                    previous_associated_router = old.external_ids.get(
                        constants.OVN_DEVICE_ID_EXT_ID_KEY)
                    if (previous_associated_router and
                            associated_router != previous_associated_router and
                            previous_associated_router in
                            self.agent.ovn_local_cr_lrps):
                        return True
        except (IndexError, AttributeError):
            return False
        return False

    def _run(self, event, row, old):
        with _SYNC_STATE_LOCK.read_lock():
            ips = row.external_ids.get(constants.OVN_CIDRS_EXT_ID_KEY,
                                       "").split()
            if event == self.ROW_DELETE:
                subnet_info = {
                    'associated_router': row.external_ids.get(
                        constants.OVN_DEVICE_ID_EXT_ID_KEY),
                    'network': common_utils.get_from_external_ids(
                        row, constants.OVN_LS_NAME_EXT_ID_KEY),
                    'address_scopes': driver_utils.get_addr_scopes(row)}
            else:
                associated_router = row.external_ids.get(
                    constants.OVN_DEVICE_ID_EXT_ID_KEY)
                if hasattr(old, 'external_ids'):
                    previous_associated_router = old.external_ids.get(
                        constants.OVN_DEVICE_ID_EXT_ID_KEY)
                    if previous_associated_router != associated_router:
                        associated_router = previous_associated_router
                subnet_info = {
                    'associated_router': associated_router,
                    'network': common_utils.get_from_external_ids(
                        row, constants.OVN_LS_NAME_EXT_ID_KEY),
                    'address_scopes': driver_utils.get_addr_scopes(row)}
            self.agent.withdraw_subnet(ips, subnet_info)


class LogicalSwitchPortTenantCreateEvent(base_watcher.LSPChassisEvent):
    def __init__(self, bgp_agent):
        events = (self.ROW_UPDATE,)
        super(LogicalSwitchPortTenantCreateEvent, self).__init__(
            bgp_agent, events)

    def match_fn(self, event, row, old):
        try:
            # single and dual-stack format
            try:
                port_utils.get_ips_from_lsp(row)
            except exceptions.IpAddressNotFound:
                return False

            if not bool(row.up[0]):
                return False

            current_network = common_utils.get_from_external_ids(
                row, constants.OVN_LS_NAME_EXT_ID_KEY)
            if current_network not in self.agent.ovn_local_lrps:
                return False

            if hasattr(old, 'up'):
                if not bool(old.up[0]):
                    return True

            if hasattr(old, 'external_ids'):
                old_network = common_utils.get_from_external_ids(
                    old, constants.OVN_LS_NAME_EXT_ID_KEY)
                if old_network != current_network:
                    return True
        except (IndexError, AttributeError):
            return False
        return False

    def _run(self, event, row, old):
        if row.type not in [constants.OVN_VM_VIF_PORT_TYPE,
                            constants.OVN_VIRTUAL_VIF_PORT_TYPE]:
            return
        with _SYNC_STATE_LOCK.read_lock():
            ips = port_utils.get_ips_from_lsp(row)
            mac = port_utils.get_mac_from_lsp(row)
            ips_info = {
                'mac': mac,
                'cidrs': row.external_ids.get(constants.OVN_CIDRS_EXT_ID_KEY,
                                              "").split(),
                'type': row.type,
                'logical_switch': common_utils.get_from_external_ids(
                    row, constants.OVN_LS_NAME_EXT_ID_KEY),
            }
            self.agent.expose_remote_ip(ips, ips_info)


class LogicalSwitchPortTenantDeleteEvent(base_watcher.LSPChassisEvent):
    def __init__(self, bgp_agent):
        events = (self.ROW_UPDATE, self.ROW_DELETE,)
        super(LogicalSwitchPortTenantDeleteEvent, self).__init__(
            bgp_agent, events)

    def match_fn(self, event, row, old):
        try:
            # single and dual-stack format
            if not port_utils.has_ip_address_defined(row.addresses[0]):
                return False

            current_network = common_utils.get_from_external_ids(
                row, constants.OVN_LS_NAME_EXT_ID_KEY)
            # Assuming the current_network cannot be changed at once
            if current_network not in self.agent.ovn_local_lrps:
                return False

            if event == self.ROW_DELETE:
                return bool(row.up[0])

            # ROW UPDATE EVENT
            if hasattr(old, 'up'):
                return bool(old.up[0])
        except (IndexError, AttributeError):
            return False
        return False

    def _run(self, event, row, old):
        if row.type not in [constants.OVN_VM_VIF_PORT_TYPE,
                            constants.OVN_VIRTUAL_VIF_PORT_TYPE]:
            return
        with _SYNC_STATE_LOCK.read_lock():
            # The address is present because of the
            # has_ip_address_defined() check in match_fn(), therefore
            # there is no need for the try-except block.
            ips = port_utils.get_ips_from_lsp(row)
            mac = port_utils.get_mac_from_lsp(row)
            ips_info = {
                'mac': mac,
                'cidrs': row.external_ids.get(constants.OVN_CIDRS_EXT_ID_KEY,
                                              "").split(),
                'type': row.type,
                'logical_switch': common_utils.get_from_external_ids(
                    row, constants.OVN_LS_NAME_EXT_ID_KEY),
            }
            self.agent.withdraw_remote_ip(ips, ips_info)


class OVNLBCreateEvent(base_watcher.OVNLBEvent):
    def __init__(self, bgp_agent):
        events = (self.ROW_UPDATE,)
        super(OVNLBCreateEvent, self).__init__(
            bgp_agent, events)

    def match_fn(self, event, row, old):
        # The ovn lb balancers are exposed through the cr-lrp, so if the
        # local agent does not have the matching router there is no need
        # to process the event
        try:
            if not row.vips:
                return False
            lb_router = router_utils.get_name_from_external_ids(row)
            if lb_router not in self.agent.ovn_local_cr_lrps.keys():
                return False

            # Expose if there is a modification in the VIPS, first new item (
            # that could happend with and non existing vips on old event or
            # empty one) or additional items because a bigger row.vips is
            # including old.vips
            if hasattr(old, 'vips'):
                if ((not old.vips and row.vips) or
                        (old.vips != row.vips and
                         set(old.vips.keys()).issubset(set(row.vips.keys())))):
                    return True

            if hasattr(old, 'external_ids'):
                # Check if the lb_router was added
                old_lb_router = router_utils.get_name_from_external_ids(old)
                if lb_router != old_lb_router:
                    return True
        except AttributeError:
            return False
        return False

    def _run(self, event, row, old):
        # vips field grows
        diff = lb_utils.get_diff_ip_from_vips(row, old)
        for ip in diff:
            with _SYNC_STATE_LOCK.read_lock():
                if lb_utils.is_vip(row, ip):
                    self.agent.expose_ovn_lb_vip(row)
                elif lb_utils.is_fip(row, ip):
                    self.agent.expose_ovn_lb_fip(row)

        # router set ext-gw
        # NOTE(froyo): Not needed to check/call to expose_ovn_lb_fip, since up
        # to this point this LB could not have been associated with a FIP
        # since the subnet did not have access to the public network
        if hasattr(old, 'external_ids'):
            with _SYNC_STATE_LOCK.read_lock():
                if (router_utils.get_name_from_external_ids(old) !=
                        router_utils.get_name_from_external_ids(row)):
                    self.agent.expose_ovn_lb_vip(row)


class OVNLBDeleteEvent(base_watcher.OVNLBEvent):
    def __init__(self, bgp_agent):
        events = (self.ROW_DELETE, self.ROW_UPDATE)
        super(OVNLBDeleteEvent, self).__init__(
            bgp_agent, events)

    def match_fn(self, event, row, old):
        # The ovn lb balancers are exposed through the cr-lrp, so if the
        # local agent does not have the matching router there is no need
        # to process the event
        try:
            if event == self.ROW_DELETE:
                if not row.vips:
                    return False
                lb_router = router_utils.get_name_from_external_ids(row)
                if lb_router in self.agent.ovn_local_cr_lrps.keys():
                    return True
                return False

            # ROW UPDATE EVENT
            lb_router = router_utils.get_name_from_external_ids(row)
            if hasattr(old, 'external_ids'):
                old_lb_router = router_utils.get_name_from_external_ids(old)
                if not old_lb_router:
                    return False
                if old_lb_router not in self.agent.ovn_local_cr_lrps.keys():
                    return False
                if old_lb_router != lb_router:
                    # Router should not be removed, but if that is the case we
                    # should remove the loadbalancer
                    return True

            # Whatever the change removing any field from vips should be manage
            if hasattr(old, 'vips'):
                if ((old.vips != row.vips and
                        set(row.vips.keys()).issubset(
                            set(old.vips.keys())))):
                    return True
        except AttributeError:
            return False
        return False

    def _run(self, event, row, old):
        # DELETE event need drop all
        if event == self.ROW_DELETE:
            diff = lb_utils.get_vips(row)
            for ip in diff:
                with _SYNC_STATE_LOCK.read_lock():
                    if lb_utils.is_vip(row, ip):
                        self.agent.withdraw_ovn_lb_vip(row)
                    elif lb_utils.is_fip(row, ip):
                        self.agent.withdraw_ovn_lb_fip(row)
            return

        # UPDATE event
        # vips field decrease
        diff = lb_utils.get_diff_ip_from_vips(old, row)
        for ip in diff:
            with _SYNC_STATE_LOCK.read_lock():
                if lb_utils.is_vip(old, ip):
                    self.agent.withdraw_ovn_lb_vip(old)
                elif lb_utils.is_fip(old, ip):
                    self.agent.withdraw_ovn_lb_fip(old)

        # router unset ext-gw
        if hasattr(old, 'external_ids'):
            with _SYNC_STATE_LOCK.read_lock():
                if (router_utils.get_name_from_external_ids(old) !=
                        router_utils.get_name_from_external_ids(row)):
                    self.agent.withdraw_ovn_lb_vip(old)


class OVNPFBaseEvent(base_watcher.OVNLBEvent):

    event = None

    def __init__(self, bgp_agent):
        super(OVNPFBaseEvent, self).__init__(
            bgp_agent, (self.event,))

    def match_fn(self, event, row, old):
        # The ovn port forwarding are manage as OVN lb balancers and they are
        # exposed through the cr-lrp, so if the local agent does not have the
        # matching router there is no need to process the event
        if not driver_utils.check_name_prefix(row,
                                              constants.OVN_LB_PF_NAME_PREFIX):
            return False

        if not row.vips:
            return False
        lb_router = router_utils.get_name_from_external_ids(
            row, constants.OVN_LR_NAME_EXT_ID_KEY)
        return lb_router in self.agent.ovn_local_cr_lrps.keys()


class OVNPFCreateEvent(OVNPFBaseEvent):

    event = OVNPFBaseEvent.ROW_CREATE

    def _run(self, event, row, old):
        with _SYNC_STATE_LOCK.read_lock():
            self.agent.expose_ovn_pf_lb_fip(row)


class OVNPFDeleteEvent(OVNPFBaseEvent):

    event = OVNPFBaseEvent.ROW_DELETE

    def _run(self, event, row, old):
        with _SYNC_STATE_LOCK.read_lock():
            self.agent.withdraw_ovn_pf_lb_fip(row)


class NATMACAddedEvent(base_watcher.DnatSnatBaseEvent):
    events = (base_watcher.DnatSnatBaseEvent.ROW_UPDATE,)

    def match_fn(self, event, row, old):
        try:
            lsp_id = row.logical_port[0]
        except IndexError:
            LOG.error("NAT entry %s has no logical port set.", row.uuid)
            return False

        lsp = self.agent.nb_idl.lsp_get(lsp_id).execute()

        if lsp is None:
            LOG.error("Logical Switch Port %(lsp)s for NAT entry %(nat)s "
                      "was not found in OVN NB DB.", {
                          'lsp': lsp_id,
                          'nat': row.uuid})
            return False

        if lsp.type == constants.OVN_VIRTUAL_VIF_PORT_TYPE:
            # The host it should be exposed to is set in the external_ids
            lsp_host = lsp.external_ids.get(constants.OVN_HOST_ID_EXT_ID_KEY)
            if lsp_host != self.agent.chassis:
                LOG.debug('NAT LSP port is bound to host %s', lsp_host)
                return False

        elif lsp.type != constants.OVN_VM_VIF_PORT_TYPE:
            LOG.debug("Logical Switch Port %(lsp)s for NAT entry %(nat)s "
                      "is not a VM port.", {'lsp': lsp_id, 'nat': row.uuid})
            return False
        else:
            try:
                if lsp.options['requested-chassis'] != self.agent.chassis:
                    return False
            except KeyError:
                return False

        try:
            if old.external_mac and old.external_mac[0] == row.external_mac[0]:
                return False
        except (AttributeError, IndexError):
            return False

        # This is required to be able to expose the FIP, there is no point in
        # continuing if the external_id is not set
        if constants.OVN_FIP_NET_EXT_ID_KEY not in row.external_ids:
            LOG.error("NAT entry %(nat)s does not have %(fip_net)s set in "
                      "external_ids", {
                          'nat': row.uuid,
                          'fip_net': constants.OVN_FIP_NET_EXT_ID_KEY})
            return False

        if not row.external_ip:
            LOG.error("NAT entry %s does not have external_ip set", row.uuid)
            return False

        return True

    def run(self, event, row, old):
        lsp = self.agent.nb_idl.lsp_get(row.logical_port[0]).execute()
        # The key is present, it was checked in match_fn
        net_id = row.external_ids[constants.OVN_FIP_NET_EXT_ID_KEY]
        ls_name = "neutron-{}".format(net_id)
        with _SYNC_STATE_LOCK.read_lock():
            self.agent.expose_fip(
                row.external_ip, row.external_mac[0], ls_name, lsp)


class ExposeFIPOnCRLRP(base_watcher.FipOnCRLRPBaseEvent):
    """Expose floating IP on the gateway chassis hosting the gateway port.

    This event happens when NAT entry is created. It exposes the floating IP on
    the gateway chassis hosting the gateway router.
    """
    events = (base_watcher.DnatSnatBaseEvent.ROW_CREATE,)

    def run(self, event, row, old):
        with _SYNC_STATE_LOCK.read_lock():
            self.agent.nat_exposer.expose_fip_from_nat(row)


class WithdrawFIPOnCRLRP(base_watcher.FipOnCRLRPBaseEvent):
    """Withdraw floating IP from the gateway chassis hosting the gateway port.

    This event happens when NAT entry is deleted. It withdraws the floating IP
    from the gateway chassis hosting the gateway router.
    """
    events = (base_watcher.DnatSnatBaseEvent.ROW_DELETE,)

    def run(self, event, row, old):
        self.agent.nat_exposer.withdraw_fip_from_nat(row)


class CrLrpChassisChangeBaseEvent(base_watcher.LRPChassisEvent):
    """Base class for case when gateway port moves.

    It matches if the hosting-chassis in status column of the gateway port has
    changed.
    """
    def __init__(self, bgp_agent):
        super().__init__(bgp_agent, (self.ROW_UPDATE,))

    def match_fn(self, event, row, old):
        new_chassis = row.status.get(constants.OVN_STATUS_CHASSIS)
        try:
            old_chassis = old.status.get(constants.OVN_STATUS_CHASSIS)
        except AttributeError:
            return False

        # Match only if the port was moved
        return new_chassis != old_chassis


class CrLrpChassisChangeExposeEvent(CrLrpChassisChangeBaseEvent):
    """A LRP event to expose floating IPs on centralized node.

    Expose all floating IPs hosted by a router with this gateway port hosted on
    this chassis. It matches in case of gateway port changes its hosting
    chassis to this chassis.
    """
    def match_fn(self, event, row, old):
        if not super().match_fn(event, row, old):
            return False

        if constants.OVN_STATUS_CHASSIS not in row.status:
            return False

        if row.status[constants.OVN_STATUS_CHASSIS] != self.agent.chassis_id:
            return False

        return True

    def run(self, event, row, old):
        nats = self.agent.nb_idl.get_nats_by_lrp(row)
        with _SYNC_STATE_LOCK.read_lock():
            for nat in nats:
                self.agent.nat_exposer.expose_fip_from_nat(nat)


class CrLrpChassisChangeWithdrawEvent(CrLrpChassisChangeBaseEvent):
    """A LRP event to withdraw floating IPs from centralized node.

    Withdraw all floating IPs hosted by a router with this gateway port hosted
    on this chassis. It matches in case of gateway port changes its hosting
    chassis from this chassis to a new one.
    """
    def match_fn(self, event, row, old):
        if not super().match_fn(event, row, old):
            return False

        # if old does not have status, it would have failed in
        # super().match_fn()
        if constants.OVN_STATUS_CHASSIS not in old.status:
            return False

        if old.status[constants.OVN_STATUS_CHASSIS] != self.agent.chassis_id:
            return False

        return True

    def run(self, event, row, old):
        nats = self.agent.nb_idl.get_nats_by_lrp(row)
        with _SYNC_STATE_LOCK.read_lock():
            for nat in nats:
                self.agent.nat_exposer.withdraw_fip_from_nat(nat)


class DistributedFlagChangedEvent(base_watcher.Event):
    """Re-register events if Neutron changed the distributed flag.

    The event matches if distributed flag was switched in OVN. Then it
    re-registers the events to react on the right events and does a full
    re-sync to withdraw or expose IPs.
    """
    def __init__(self, bgp_agent):
        table = 'NB_Global'
        events = (self.ROW_UPDATE,)
        super().__init__(bgp_agent, events, table)
        self.event_name = self.__class__.__name__

    def match_fn(self, event, row, old):
        try:
            if (old.external_ids.get(constants.OVN_FIP_DISTRIBUTED) ==
                    row.external_ids[constants.OVN_FIP_DISTRIBUTED]):
                return False
        except KeyError:
            # Distributed flag was deleted, behave like distributed agent
            pass
        except AttributeError:
            return False

        return True

    def run(self, event, row, old):
        if row.external_ids.get(constants.OVN_FIP_DISTRIBUTED) == "True":
            self.agent.distributed = True
        elif row.external_ids.get(constants.OVN_FIP_DISTRIBUTED) == "False":
            self.agent.distributed = False
        else:
            # Default to True
            self.agent.distributed = True

        self.agent.sync()
        self.agent.frr_sync()



# Copyright 2021 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import dataclasses
import ipaddress
import threading

from oslo_concurrency import lockutils
from oslo_config import cfg
from oslo_log import log as logging

from ovn_bgp_agent import constants
from ovn_bgp_agent.drivers import driver_api
from ovn_bgp_agent.drivers.openstack.utils import bgp as bgp_utils
from ovn_bgp_agent.drivers.openstack.utils import driver_utils
from ovn_bgp_agent.drivers.openstack.utils import frr
from ovn_bgp_agent.drivers.openstack.utils import ovn
from ovn_bgp_agent.drivers.openstack.utils import ovs
from ovn_bgp_agent.drivers.openstack.watchers import bgp_watcher as watcher
from ovn_bgp_agent import exceptions as agent_exc
from ovn_bgp_agent.utils import linux_net


CONF = cfg.CONF
LOG = logging.getLogger(__name__)

OVN_TABLES = ["Port_Binding", "Chassis", "Datapath_Binding", "Chassis_Private"]


@dataclasses.dataclass(frozen=True, eq=True)
class HashedRoute:
    network: str
    prefix_len: int
    dst: str


class OVNBGPStretchedL2Driver(driver_api.AgentDriverBase):
    def __init__(self):
        self.ovn_local_cr_lrps = {}
        self.vrf_routes = set()
        self.ovn_routing_tables_routes = collections.defaultdict()
        self.allowed_address_scopes = set(CONF.address_scopes or [])
        self.propagated_lrp_ports = {}

        self._sb_idl = None
        self._post_fork_event = threading.Event()

    @property
    def sb_idl(self):
        if not self._sb_idl:
            self._post_fork_event.wait()
        return self._sb_idl

    @sb_idl.setter
    def sb_idl(self, val):
        self._sb_idl = val

    def start(self):
        self.ovs_idl = ovs.OvsIdl()
        self.ovs_idl.start(CONF.ovsdb_connection)

        # Base BGP configuration
        # Ensure FRR is configured to leak only kernel routes by default
        frr.set_default_redistribute(['kernel'])
        bgp_utils.ensure_base_bgp_configuration()

        # Clear vrf routing table
        if CONF.clear_vrf_routes_on_startup:
            linux_net.delete_routes_from_table(CONF.bgp_vrf_table_id)

        self.chassis = self.ovs_idl.get_own_chassis_id()
        self.ovn_remote = self.ovs_idl.get_ovn_remote()
        LOG.debug("Loaded chassis %s.", self.chassis)
        if self.allowed_address_scopes:
            LOG.debug("Configured allowed address scopes: %s",
                      ", ".join(self.allowed_address_scopes))

        self._post_fork_event.clear()

        events = self._get_events()
        self.sb_idl = ovn.OvnSbIdl(
            self.ovn_remote,
            chassis=self.chassis,
            tables=OVN_TABLES,
            events=events,
        ).start()

        # Now IDL connections can be safely used
        self._post_fork_event.set()

    def _get_events(self):
        return {
            watcher.SubnetRouterAttachedEvent(self),
            watcher.SubnetRouterUpdateEvent(self),
            watcher.SubnetRouterDetachedEvent(self),
            watcher.PortBindingChassisCreatedEvent(self),
            watcher.PortBindingChassisDeletedEvent(self),
        }

    @lockutils.synchronized('bgp')
    def frr_sync(self):
        LOG.debug("Ensuring VRF configuration for advertising routes")
        # Base BGP configuration
        # Ensure FRR is configured to leak the routes
        bgp_utils.ensure_base_bgp_configuration()

    @lockutils.synchronized("bgp")
    def sync(self):
        self.ovn_local_cr_lrps = {}
        self.ovn_routing_tables_routes = collections.defaultdict()
        self.vrf_routes = set()
        self.propagated_lrp_ports = {}

        LOG.debug("Syncing current routes.")

        # Get all current exposed routes
        vrf_routes = linux_net.get_routes_on_tables([CONF.bgp_vrf_table_id])

        for cr_lrp_port in self.sb_idl.get_cr_lrp_ports():
            if (not cr_lrp_port.mac or
                    len(cr_lrp_port.mac[0].strip().split(" ")) <= 1):
                continue

            self._expose_cr_lrp(cr_lrp_port.mac[0].strip().split(" ")[1:],
                                cr_lrp_port)

        # remove all left over routes
        delete_routes = []
        for route in vrf_routes:
            r = HashedRoute(
                network=route.dst,
                prefix_len=route.dst_len,
                dst=route.gateway if route.gateway else None)
            if r not in self.vrf_routes:
                delete_routes.append(route)

        linux_net.delete_ip_routes(delete_routes)

    def _add_route(self, network, prefix_len, dst=None):
        LOG.debug("Adding BGP route for Network %s/%d via %s",
                  network, prefix_len, dst)

        linux_net.add_ip_route(
            self.ovn_routing_tables_routes,
            network,
            CONF.bgp_vrf_table_id,
            CONF.bgp_nic,
            vlan=None,
            mask=prefix_len,
            via=dst)
        r = HashedRoute(
            network=network,
            prefix_len=prefix_len,
            dst=dst)
        self.vrf_routes.add(r)

        LOG.debug("Added BGP route for Network %s/%d via %s",
                  network, prefix_len, dst)

    def _del_route(self, network, prefix_len, dst=None):
        LOG.debug("Deleting BGP route for Network %s/%d via %s",
                  network, prefix_len, dst)

        linux_net.del_ip_route(
            self.ovn_routing_tables_routes,
            network,
            CONF.bgp_vrf_table_id,
            CONF.bgp_nic,
            vlan=None,
            mask=prefix_len,
            via=dst)
        r = HashedRoute(
            network=network,
            prefix_len=prefix_len,
            dst=dst)
        if r in self.vrf_routes:
            self.vrf_routes.remove(r)

        LOG.debug("Deleted BGP route for Network %s/%d via %s",
                  network, prefix_len, dst)

    def _address_scope_allowed(self, scope1, scope2, ip_version):
        if not self.allowed_address_scopes:
            # No address scopes to filter on => announce everything
            return True

        if scope1[ip_version] != scope2[ip_version]:
            # Not the same address scope => don't announce
            return False

        if scope1[ip_version] not in self.allowed_address_scopes:
            # This address scope does not match => don't announce
            return False

        return True

    @lockutils.synchronized("bgp")
    def expose_subnet(self, ip, row):
        try:
            cr_lrp = self.sb_idl.is_router_gateway_on_any_chassis(row.datapath)
        except agent_exc.DatapathNotFound:
            LOG.debug("Port %s not being exposed as its datapath %s was "
                      "removed", row.logical_port, row.datapath)
            return
        if not cr_lrp:
            return

        self._ensure_network_exposed(row, cr_lrp.logical_port)

    @lockutils.synchronized("bgp")
    def update_subnet(self, old, row):
        try:
            cr_lrp = self.sb_idl.is_router_gateway_on_any_chassis(row.datapath)
        except agent_exc.DatapathNotFound:
            LOG.debug("Port %s not being updated as its datapath %s was "
                      "removed", row.logical_port, row.datapath)
            return
        if (not cr_lrp or not cr_lrp.mac or
                len(cr_lrp.mac[0].strip().split(" ")) <= 1):
            return

        current_ips = row.mac[0].strip().split(" ")[1:]
        previous_ips = (
            old.mac[0].strip().split(" ")[1:]
            if old.mac or len(old.mac[0].strip().split(" ")) > 1
            else []
        )
        add_ips = list(
            filter(lambda ip: ip not in previous_ips, current_ips))
        delete_ips = list(
            filter(lambda ip: ip not in current_ips, previous_ips))

        self._update_network(row, cr_lrp.logical_port, add_ips, delete_ips)

    @lockutils.synchronized("bgp")
    def withdraw_subnet(self, ip, row):
        port_info = self.propagated_lrp_ports.get(row.logical_port)
        if not port_info:
            return

        self._withdraw_subnet(port_info, port_info["cr_lrp"])

        gateway = self.ovn_local_cr_lrps.get(port_info["cr_lrp"])
        if gateway and row.logical_port in gateway["lrp_ports"]:
            gateway["lrp_ports"].remove(row.logical_port)
        self.propagated_lrp_ports.pop(row.logical_port)

    def _withdraw_subnet(self, port_info, cr_lrp):
        gateway = self.ovn_local_cr_lrps.get(cr_lrp)
        if not gateway:
            # If we dont have it cached then its either not existing or
            # or we got an event while starting up which then the sync
            # function can fix.
            return
        gateway_ips = gateway["ips"]

        subnets = [
            ipaddress.ip_network(subnet)
            for subnet in port_info["subnets"]]

        for gateway_ip in gateway_ips:
            for subnet in subnets:
                if gateway_ip.version != subnet.version:
                    continue

                self._del_route(
                    network=str(subnet.network_address),
                    prefix_len=subnet.prefixlen,
                    dst=str(gateway_ip.ip))

            # Check if can delete the link-local route
            exposed_routes = linux_net.get_exposed_routes_on_network(
                [CONF.bgp_vrf_table_id],
                gateway_ip.network)

            if not exposed_routes:
                self._del_route(
                    network=str(gateway_ip.network.network_address),
                    prefix_len=gateway_ip.network.prefixlen)

    @lockutils.synchronized('bgp')
    def withdraw_ip(self, ips, row, associated_port=None):
        if not (row.type == constants.OVN_CHASSISREDIRECT_VIF_PORT_TYPE and
                row.logical_port.startswith("cr-")):
            return
        self._withdraw_cr_lrp(ips, row)

    def _withdraw_cr_lrp(self, ips, row):
        if self.allowed_address_scopes:
            # Validate address scopes
            address_scopes = self.ovn_local_cr_lrps[row.logical_port][
                "address_scopes"]
            if not any([
                    scope in self.allowed_address_scopes
                    for scope in address_scopes.values()]):
                return

        # Check if there are networks attached to the router,
        # and if so, remove them locally
        lrp_ports = self.ovn_local_cr_lrps[row.logical_port]["lrp_ports"]
        for lrp_logical_port in lrp_ports:
            port_info = self.propagated_lrp_ports.get(lrp_logical_port)
            if not port_info:
                continue
            # withdraw network
            self._withdraw_subnet(port_info, row.logical_port)
            self.propagated_lrp_ports.pop(lrp_logical_port, None)

        self.ovn_local_cr_lrps.pop(row.logical_port, None)

    @lockutils.synchronized("bgp")
    def expose_ip(self, ips, row, associated_port=None):
        if not (row.type == constants.OVN_CHASSISREDIRECT_VIF_PORT_TYPE and
                row.logical_port.startswith("cr-")):
            return
        self._expose_cr_lrp(ips, row)

    def _expose_cr_lrp(self, ips, row):
        LOG.debug("Adding BGP route for CR-LRP Port %s", row.logical_port)
        # Keeping information about the associated network for
        # tenant network advertisement
        self.ovn_local_cr_lrps[row.logical_port] = {
            "ips": [ipaddress.ip_interface(ip) for ip in ips],
            "address_scopes": {},
            "lrp_ports": set(),
        }

        if self.allowed_address_scopes:
            # Validate address scopes
            patch_port = row.logical_port.split("cr-lrp-")[1]
            port = self.sb_idl.get_port_by_name(patch_port)
            if not port:
                LOG.error("Patchport %s for CR-LRP %s missing, skipping.",
                          patch_port, row.logical_port)
                return
            address_scopes = driver_utils.get_addr_scopes(port)
            self.ovn_local_cr_lrps[row.logical_port][
                "address_scopes"] = address_scopes
            if not any([
                    scope in self.allowed_address_scopes
                    for scope in address_scopes.values()]):
                return

        # Check if there are networks attached to the router,
        # and if so, add the needed routes
        lrp_ports = self.sb_idl.get_lrp_ports_for_router(row.datapath)
        for lrp in lrp_ports:
            if (
                lrp.chassis or
                not lrp.logical_port.startswith("lrp-") or
                "chassis-redirect-port" in lrp.options.keys()
            ):
                continue
            # expose network
            self._ensure_network_exposed(lrp, row.logical_port)

    def _update_network(self, router_port, gateway_port, add_ips, delete_ips):
        gateway = self.ovn_local_cr_lrps.get(gateway_port)
        if not gateway:
            # If we dont have it cached then its either not existing or
            # or we got an event while starting up which then the sync
            # function can fix.
            return
        gateway_ips = gateway["ips"]
        if (not router_port.mac or
                len(router_port.mac[0].strip().split(" ")) <= 1):
            return

        # get all ips from the router port
        ips_to_add = [ipaddress.ip_interface(ip) for ip in add_ips]

        ips_to_delete = [ipaddress.ip_interface(ip) for ip in delete_ips]

        for router_ip in ips_to_add + ips_to_delete:
            if router_ip in gateway_ips:
                return

        address_scopes = None
        if self.allowed_address_scopes:
            patch_port = router_port.logical_port.split("lrp-")[1]
            port = self.sb_idl.get_port_by_name(patch_port)
            if not port:
                LOG.error("Patchport %s for CR-LRP %s missing, skipping.",
                          patch_port, gateway_port)
                return
            address_scopes = driver_utils.get_addr_scopes(port)
            # if we should filter on address scopes and this port has no
            # address scopes set we do not need to go further
            if not any(address_scopes.values()):
                return

        subnets = set()
        for gateway_ip in gateway_ips:
            for router_ip in ips_to_add:
                if gateway_ip.version != router_ip.version:
                    continue

                if not self._address_scope_allowed(
                    gateway["address_scopes"],
                    address_scopes,
                    router_ip.version):
                    continue

                # Add link-local route
                self._add_route(
                    network=str(gateway_ip.network.network_address),
                    prefix_len=gateway_ip.network.prefixlen)

                # add route for the tenant network pointing to the
                # gateway ip
                self._add_route(
                    network=str(router_ip.network.network_address),
                    prefix_len=router_ip.network.prefixlen,
                    dst=str(gateway_ip.ip))
                subnets.add(str(router_ip.network))

            for router_ip in ips_to_delete:
                if gateway_ip.version != router_ip.version:
                    continue

                if not self._address_scope_allowed(
                    gateway["address_scopes"],
                    address_scopes,
                    router_ip.version):
                    continue

                self._del_route(
                    network=str(router_ip.network.network_address),
                    prefix_len=router_ip.network.prefixlen,
                    dst=str(gateway_ip.ip))

            # We only need to check this if we really deleted a route for
            # a tenant network
            if ips_to_delete:
                # Check if can delete the link-local route
                exposed_routes = linux_net.get_exposed_routes_on_network(
                    [CONF.bgp_vrf_table_id],
                    gateway_ip.network
                )

                if not exposed_routes:
                    self._del_route(
                        network=str(gateway_ip.network.network_address),
                        prefix_len=gateway_ip.network.prefixlen)

        self.ovn_local_cr_lrps[gateway_port]["lrp_ports"].add(
            router_port.logical_port)
        self.propagated_lrp_ports[router_port.logical_port] = {
            "cr_lrp": gateway_port,
            "subnets": subnets
        }

    def _ensure_network_exposed(self, router_port, gateway_port):
        gateway = self.ovn_local_cr_lrps.get(gateway_port)
        if not gateway:
            # If we dont have it cached then its either not existing or
            # or we got an event while starting up which then the sync
            # function can fix.
            return
        gateway_ips = gateway["ips"]
        if (not router_port.mac or
                len(router_port.mac[0].strip().split(" ")) <= 1):
            return

        # get all ips from the router port
        router_ips = [
            ipaddress.ip_interface(ip)
            for ip in router_port.mac[0].strip().split(" ")[1:]]

        for router_ip in router_ips:
            if router_ip in gateway_ips:
                return

        address_scopes = None
        if self.allowed_address_scopes:
            patch_port = router_port.logical_port.split("lrp-")[1]
            port = self.sb_idl.get_port_by_name(patch_port)
            if not port:
                LOG.error("Patchport %s for CR-LRP %s missing, skipping.",
                          patch_port, gateway_port)
                return
            address_scopes = driver_utils.get_addr_scopes(port)
            # if we have address scopes configured and none of them matches
            # for this port, we can skip further processing
            if not any(address_scopes.values()):
                return

        subnets = set()
        for gateway_ip in gateway_ips:
            for router_ip in router_ips:
                if gateway_ip.version != router_ip.version:
                    continue

                if not self._address_scope_allowed(
                    gateway["address_scopes"],
                    address_scopes,
                    router_ip.version):
                    continue

                # Add link-local route
                self._add_route(
                    network=str(gateway_ip.network.network_address),
                    prefix_len=gateway_ip.network.prefixlen)

                # add route for the tenant network pointing to the
                # gateway ip
                self._add_route(
                    network=str(router_ip.network.network_address),
                    prefix_len=router_ip.network.prefixlen,
                    dst=str(gateway_ip.ip))
                subnets.add(str(router_ip.network))

        if subnets:
            self.ovn_local_cr_lrps[gateway_port]["lrp_ports"].add(
                router_port.logical_port)
            self.propagated_lrp_ports[router_port.logical_port] = {
                "cr_lrp": gateway_port,
                "subnets": subnets
            }

    @lockutils.synchronized("bgp")
    def expose_remote_ip(self, ip_address):
        raise NotImplementedError()

    @lockutils.synchronized("bgp")
    def withdraw_remote_ip(self, ip_address):
        raise NotImplementedError()



# Copyright 2022 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_concurrency import processutils

from ovn_bgp_agent import constants
from ovn_bgp_agent.privileged import linux_net as priv_linux_net
from ovn_bgp_agent.tests import base as test_base
from ovn_bgp_agent.utils import linux_net


class FakeException(Exception):
    stderr = ''


class TestPrivilegedLinuxNet(test_base.TestCase):

    def setUp(self):
        super(TestPrivilegedLinuxNet, self).setUp()
        # Mock pyroute2.NDB context manager object
        self.mock_ndb = mock.patch.object(linux_net.pyroute2, 'NDB').start()
        self.fake_ndb = self.mock_ndb().__enter__()
        # Mock pyroute2.IPRoute context manager object
        self.mock_iproute = mock.patch.object(
            linux_net.pyroute2, 'IPRoute').start()
        self.fake_iproute = self.mock_iproute().__enter__()

        self.mock_exc = mock.patch.object(processutils, 'execute').start()

        # Helper variables used accross many tests
        self.ip = '10.10.1.16'
        self.ipv6 = '2002::1234:abcd:ffff:c0a8:101'
        self.dev = 'ethfake'
        self.dev_br = 'fake-provider'
        self.mac = 'aa:bb:cc:dd:ee:ff'

    def test_set_kernel_flag(self):
        priv_linux_net.set_kernel_flag('net.ipv6.conf.fake', 1)
        self.mock_exc.assert_called_once_with(
            'sysctl', '-w', 'net.ipv6.conf.fake=1')

    def test_set_kernel_flag_exception(self):
        self.mock_exc.side_effect = FakeException()
        self.assertRaises(
            FakeException,
            priv_linux_net.set_kernel_flag, 'net.ipv6.conf.fake', 1)

    def test_add_ndp_proxy(self):
        priv_linux_net.add_ndp_proxy(self.ipv6, self.dev)
        self.mock_exc.assert_called_once_with(
            'ip', '-6', 'nei', 'add', 'proxy', self.ipv6, 'dev', self.dev)

    def test_add_ndp_proxy_vlan(self):
        priv_linux_net.add_ndp_proxy(self.ipv6, self.dev, vlan=10)
        self.mock_exc.assert_called_once_with(
            'ip', '-6', 'nei', 'add', 'proxy', self.ipv6,
            'dev', '%s.10' % self.dev)

    def test_add_trimmed_ndp_proxy_vlan(self):
        priv_linux_net.add_ndp_proxy(self.ipv6, self.dev_br, vlan=1024)
        self.mock_exc.assert_called_once_with(
            'ip', '-6', 'nei', 'add', 'proxy', self.ipv6,
            'dev', '%s.1024' %
            self.dev_br[:constants.OVN_VLAN_DEVICE_MAX_LENGTH])

    def test_add_ndp_proxy_exception(self):
        self.mock_exc.side_effect = FakeException()
        self.assertRaises(
            FakeException,
            priv_linux_net.add_ndp_proxy, self.ipv6, self.dev)

    def test_del_ndp_proxy(self):
        priv_linux_net.del_ndp_proxy(self.ipv6, self.dev)
        self.mock_exc.assert_called_once_with(
            'ip', '-6', 'nei', 'del', 'proxy', self.ipv6, 'dev',
            self.dev, env_variables=mock.ANY)

    def test_del_ndp_proxy_vlan(self):
        priv_linux_net.del_ndp_proxy(self.ipv6, self.dev, vlan=10)
        self.mock_exc.assert_called_once_with(
            'ip', '-6', 'nei', 'del', 'proxy', self.ipv6, 'dev',
            '%s.10' % self.dev, env_variables=mock.ANY)

    def test_del_trimeed_ndp_proxy_vlan(self):
        priv_linux_net.del_ndp_proxy(self.ipv6, self.dev_br, vlan=1024)
        self.mock_exc.assert_called_once_with(
            'ip', '-6', 'nei', 'del', 'proxy', self.ipv6, 'dev',
            '%s.1024' % self.dev_br[:constants.OVN_VLAN_DEVICE_MAX_LENGTH],
            env_variables=mock.ANY)

    def test_del_ndp_proxy_exception(self):
        self.mock_exc.side_effect = FakeException()
        self.assertRaises(
            FakeException,
            priv_linux_net.del_ndp_proxy, self.ipv6, self.dev)

    def test_del_ndp_proxy_exception_no_such_file(self):
        exp = FakeException()
        exp.stderr = 'No such file or directory'
        self.mock_exc.side_effect = exp
        self.assertIsNone(priv_linux_net.del_ndp_proxy(self.ipv6, self.dev))

    @mock.patch('builtins.open', new_callable=mock.mock_open())
    def test_create_routing_table_for_bridge(self, mock_o):
        priv_linux_net.create_routing_table_for_bridge(17, 'fake-bridge')
        mock_o.assert_called_once_with('/etc/iproute2/rt_tables', 'a')
        mock_o().__enter__().write.assert_called_once_with('17 fake-bridge\n')



# Copyright 2023 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg
from oslo_log import log as logging

from ovn_bgp_agent import constants
from ovn_bgp_agent.drivers.openstack.utils import driver_utils
from ovn_bgp_agent.drivers.openstack.utils import evpn
from ovn_bgp_agent.drivers.openstack.utils import frr
from ovn_bgp_agent.utils import linux_net


CONF = cfg.CONF
LOG = logging.getLogger(__name__)


def announce_ips(port_ips, ips_info=None):
    if CONF.exposing_method in [constants.EXPOSE_METHOD_VRF]:
        if ips_info is None or ips_info.get('bridge_device', None) is None:
            return

        # Lookup the EVPN vlan dev
        vlan_dev = evpn.lookup_vlan(ips_info['bridge_device'],
                                    ips_info['bridge_vlan'])

        # Split the via ip's per ip version
        via = driver_utils.ips_per_version(ips_info.get('via', []))

        for ip in port_ips:
            # Add route for each ip in routing table.
            if via:
                ver = linux_net.get_ip_version(ip)
                vlan_dev.add_route(None, ip, None, via=via.get(ver))
            else:
                vlan_dev.add_route(None, ip, ips_info['mac'])
        return

    linux_net.add_ips_to_dev(CONF.bgp_nic, port_ips)


def withdraw_ips(port_ips, ips_info=None):
    if CONF.exposing_method in [constants.EXPOSE_METHOD_VRF]:
        if ips_info is None or ips_info.get('bridge_device', None) is None:
            return

        # Lookup the EVPN vlan dev
        vlan_dev = evpn.lookup_vlan(ips_info['bridge_device'],
                                    ips_info['bridge_vlan'])

        for ip in port_ips:
            # Add route for each ip in routing table.
            vlan_dev.del_route(None, ip)
        return

    linux_net.del_ips_from_dev(CONF.bgp_nic, port_ips)


def ensure_base_bgp_configuration(template=frr.LEAK_VRF_TEMPLATE):
    if CONF.exposing_method not in [constants.EXPOSE_METHOD_UNDERLAY,
                                    constants.EXPOSE_METHOD_DYNAMIC,
                                    constants.EXPOSE_METHOD_OVN]:
        return

    # Create VRF
    linux_net.ensure_vrf(CONF.bgp_vrf, CONF.bgp_vrf_table_id)

    # If we expose subnet routes, we should add kernel routes too.
    if CONF.advertisement_method_tenant_networks == 'subnet':
        frr.set_default_redistribute(['connected', 'kernel'])

    # Ensure FRR is configure to leak the routes
    frr.vrf_leak(CONF.bgp_vrf, CONF.bgp_AS, CONF.bgp_router_id,
                 template=template)

    # Create OVN dummy device
    linux_net.ensure_ovn_device(CONF.bgp_nic, CONF.bgp_vrf)



# Copyright 2024 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy

from ovsdbapp.backend.ovs_idl import event

from ovn_bgp_agent import constants
from ovn_bgp_agent.tests.functional import base
from ovn_bgp_agent.tests import utils


class DistributedWaitEvent(event.WaitEvent):
    event_name = 'DistributedWaitEvent'

    def __init__(self, timeout=5):
        table = 'NB_Global'
        events = (self.ROW_UPDATE,)
        super().__init__(events, table, None, timeout=timeout)

    def match_fn(self, event, row, old):
        return row.external_ids != getattr(old, 'external_ids')


class DistributedFlagChangedEventTestCase(
    base.BaseFunctionalNBAgentTestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.agent_config = copy.copy(cls.agent_config)
        cls.agent_config.setdefault(None, {})[
            'advertisement_method_tenant_networks'] = 'subnet'

    def _make_nb_global_event(self):
        nb_global_event = DistributedWaitEvent()
        self.nb_api.ovsdb_connection.idl.notify_handler.watch_event(
            nb_global_event)
        return nb_global_event

    def _wait_for_events_added(self, events):
        def _events_intersect():
            registered_events = set(idl.notify_handler._watched_events)
            intersection = events & registered_events
            return intersection == events

        idl = self.agent.nb_idl.ovsdb_connection.idl

        utils.wait_until_true(
            _events_intersect,
            timeout=5,
            sleep=0.1,
            exception=AssertionError(
                "Events %s still not registered in the agent" % events))

    def _wait_for_events_removed(self, events):
        def _events_disjunctive():
            registered_events = set(idl.notify_handler._watched_events)
            return not bool(events & registered_events)

        idl = self.agent.nb_idl.ovsdb_connection.idl

        utils.wait_until_true(
            _events_disjunctive,
            timeout=5,
            sleep=0.1,
            exception=AssertionError(
                "Events %s still registered in the agent" % events))

    def test_distributed_flag_changed(self):
        distributed = self.nb_api.db_get(
            'NB_Global', '.', 'external_ids').execute(check_error=True).get(
                constants.OVN_FIP_DISTRIBUTED)

        self.assertIsNone(distributed)

        distributed_events = set(self.agent._get_additional_events(
            distributed=True))
        centralized_events = set(self.agent._get_additional_events(
            distributed=False))

        # At start there is no distributed flag but the agent should default to
        # distributed
        self._wait_for_events_added(distributed_events)
        self._wait_for_events_removed(centralized_events)

        nb_global_event = self._make_nb_global_event()
        self.nb_api.db_set('NB_Global', '.', external_ids={
            constants.OVN_FIP_DISTRIBUTED: "False"}).execute(check_error=True)

        self.assertTrue(nb_global_event.wait())

        self._wait_for_events_added(centralized_events)
        self._wait_for_events_removed(distributed_events)

        nb_global_event = self._make_nb_global_event()
        self.nb_api.db_set('NB_Global', '.', external_ids={
            constants.OVN_FIP_DISTRIBUTED: "True"}).execute(check_error=True)

        self.assertTrue(nb_global_event.wait())

        self._wait_for_events_added(distributed_events)
        self._wait_for_events_removed(centralized_events)

        nb_global_event = self._make_nb_global_event()
        self.nb_api.db_set('NB_Global', '.', external_ids={
            constants.OVN_FIP_DISTRIBUTED: "False"}).execute(check_error=True)

        self.assertTrue(nb_global_event.wait())

        self._wait_for_events_added(centralized_events)
        self._wait_for_events_removed(distributed_events)

        nb_global_event = self._make_nb_global_event()
        self.nb_api.db_remove(
            'NB_Global', '.', 'external_ids',
            constants.OVN_FIP_DISTRIBUTED).execute(check_error=True)

        self.assertTrue(nb_global_event.wait())

        self._wait_for_events_added(distributed_events)
        self._wait_for_events_removed(centralized_events)



# Copyright 2021 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import tempfile

from jinja2 import Template
from oslo_config import cfg
from oslo_log import log as logging

from ovn_bgp_agent import constants
import ovn_bgp_agent.privileged.vtysh

CONF = cfg.CONF

LOG = logging.getLogger(__name__)

DEFAULT_REDISTRIBUTE = {'connected'}

CONFIGURE_ND_TEMPLATE = '''
interface {{ intf }}
{% if is_dhcpv6 %}
 ipv6 nd managed-config-flag
{% endif %}
{% for server in dns_servers %}
 ipv6 nd rdnss {{ server }}
{% endfor %}
 ipv6 nd prefix {{ prefix }}
 no ipv6 nd suppress-ra
exit
'''

ADD_VRF_TEMPLATE = '''
vrf {{ vrf_name }}
  vni {{ vni }}
exit-vrf

router bgp {{ bgp_as }} vrf {{ vrf_name }}
  address-family ipv4 unicast
{% for redist in redistribute %}
    redistribute {{ redist }}
{% endfor %}
  exit-address-family
  address-family ipv6 unicast
{% for redist in redistribute %}
    redistribute {{ redist }}
{% endfor %}
  exit-address-family
  address-family l2vpn evpn
    advertise ipv4 unicast
    advertise ipv6 unicast
{% if route_distinguishers|length > 0 %}
    rd {{ route_distinguishers[0] }}
{% else %}
    rd {{ local_ip }}:{{ vni }}
{% endif %}
{% for route_target in route_targets %}
    route-target import {{ route_target }}
    route-target export {{ route_target }}
{% endfor %}
{% for route_target in export_targets %}
    route-target export {{ route_target }}
{% endfor %}
{% for route_target in import_targets %}
    route-target import {{ route_target }}
{% endfor %}
  exit-address-family

'''

DEL_VRF_TEMPLATE = '''
no vrf {{ vrf_name }}
no interface veth-{{ vrf_name }}
no router bgp {{ bgp_as }} vrf {{ vrf_name }}

'''

LEAK_VRF_TEMPLATE = '''
router bgp {{ bgp_as }}
  address-family ipv4 unicast
    import vrf {{ vrf_name }}
  exit-address-family

  address-family ipv6 unicast
    import vrf {{ vrf_name }}
  exit-address-family

router bgp {{ bgp_as }} vrf {{ vrf_name }}
  bgp router-id {{ bgp_router_id }}
  address-family ipv4 unicast
{% for redist in redistribute %}
    redistribute {{ redist }}
{% endfor %}
  exit-address-family

  address-family ipv6 unicast
{% for redist in redistribute %}
    redistribute {{ redist }}
{% endfor %}
  exit-address-family

'''


def _get_router_id():
    output = ovn_bgp_agent.privileged.vtysh.run_vtysh_command(
        command='show ip bgp summary json')
    return json.loads(output).get('ipv4Unicast', {}).get('routerId')


def _run_vtysh_config_with_tempfile(vrf_config):
    try:
        f = tempfile.NamedTemporaryFile(mode='w')
        f.write(vrf_config)
        f.flush()
    except (IOError, OSError) as e:
        LOG.error('Failed to create the VRF configuration '
                  'file. Error: %s', e)
        if f is not None:
            f.close()
        raise

    try:
        ovn_bgp_agent.privileged.vtysh.run_vtysh_config(f.name)
    finally:
        if f is not None:
            f.close()


def set_default_redistribute(redist_opts):
    if not isinstance(redist_opts, set):
        redist_opts = set(redist_opts)

    if redist_opts == DEFAULT_REDISTRIBUTE:
        # no update required.
        return

    DEFAULT_REDISTRIBUTE.clear()
    DEFAULT_REDISTRIBUTE.update(redist_opts)


def nd_reconfigure(interface, prefix, opts):
    LOG.info('FRR IPv6 ND reconfiguration (intf %s, prefix %s)', interface,
             prefix)
    nd_template = Template(CONFIGURE_ND_TEMPLATE)

    # Need to define what setting is for SLAAC
    if (not opts.get('dhcpv6_stateless', False) or
            opts.get('dhcpv6_stateless', '') not in ('true', True)):
        prefix += ' no-autoconfig'

    # Parse dns servers from dhcp options.
    dns_servers = []
    if opts.get('dns_server'):
        dns_servers = [s.strip() for s in opts['dns_server'][1:-1].split(',')]

    is_dhcpv6 = True  # Need a better way to define this one.

    nd_config = nd_template.render(
        intf=interface,
        prefix=prefix,
        dns_servers=dns_servers,
        is_dhcpv6=is_dhcpv6,
    )

    _run_vtysh_config_with_tempfile(nd_config)


def vrf_leak(vrf, bgp_as, bgp_router_id=None, template=LEAK_VRF_TEMPLATE):
    LOG.info("Add VRF leak for VRF %s on router bgp %s", vrf, bgp_as)
    if not bgp_router_id:
        bgp_router_id = _get_router_id()
        if not bgp_router_id:
            LOG.error("Unknown router-id, needed for route leaking")
            return

    vrf_template = Template(template)
    vrf_config = vrf_template.render(vrf_name=vrf, bgp_as=bgp_as,
                                     redistribute=DEFAULT_REDISTRIBUTE,
                                     bgp_router_id=bgp_router_id)
    _run_vtysh_config_with_tempfile(vrf_config)


def vrf_reconfigure(evpn_info, action):
    LOG.info("FRR reconfiguration (action = %s) for evpn: %s",
             action, evpn_info)

    # If we have more actions, we can define them in this list.
    vrf_templates = {
        'add-vrf': ADD_VRF_TEMPLATE,
        'del-vrf': DEL_VRF_TEMPLATE,
    }
    if action not in vrf_templates:
        LOG.error("Unknown FRR reconfiguration action: %s", action)
        return

    # Set default opts, so all params are available for the templates
    # Then update them with evpn_info
    opts = dict(route_targets=[], route_distinguishers=[], export_targets=[],
                import_targets=[], local_ip=CONF.evpn_local_ip,
                redistribute=DEFAULT_REDISTRIBUTE,
                bgp_as=CONF.bgp_AS, vrf_name='', vni=0)
    opts.update(evpn_info)

    if not opts['vrf_name']:
        opts['vrf_name'] = "{}{}".format(constants.OVN_EVPN_VRF_PREFIX,
                                         evpn_info['vni'])

    vrf_template = Template(vrf_templates.get(action))
    vrf_config = vrf_template.render(**opts)

    _run_vtysh_config_with_tempfile(vrf_config)



# Copyright 2021 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_concurrency import lockutils
from oslo_log import log as logging

from ovn_bgp_agent import constants
from ovn_bgp_agent.drivers.openstack.utils import port as port_utils
from ovn_bgp_agent.drivers.openstack.watchers import base_watcher


LOG = logging.getLogger(__name__)

_SYNC_STATE_LOCK = lockutils.ReaderWriterLock()


class PortBindingChassisCreatedEvent(base_watcher.PortBindingChassisEvent):
    def __init__(self, bgp_agent):
        events = (self.ROW_UPDATE,)
        super(PortBindingChassisCreatedEvent, self).__init__(
            bgp_agent, events)

    def match_fn(self, event, row, old):
        try:
            if row.type != constants.OVN_CHASSISREDIRECT_VIF_PORT_TYPE:
                return False
            # single and dual-stack format
            if not port_utils.has_ip_address_defined(row.mac[0]):
                return False
            return (row.chassis[0].name == self.agent.chassis and
                    not old.chassis)
        except (IndexError, AttributeError):
            return False

    def _run(self, event, row, old):
        if row.type != constants.OVN_CHASSISREDIRECT_VIF_PORT_TYPE:
            return
        with _SYNC_STATE_LOCK.read_lock():
            self.agent.expose_ip(row, cr_lrp=True)


class PortBindingChassisDeletedEvent(base_watcher.PortBindingChassisEvent):
    def __init__(self, bgp_agent):
        events = (self.ROW_UPDATE, self.ROW_DELETE,)
        super(PortBindingChassisDeletedEvent, self).__init__(
            bgp_agent, events)

    def match_fn(self, event, row, old):
        try:
            if row.type != constants.OVN_CHASSISREDIRECT_VIF_PORT_TYPE:
                return False
            # single and dual-stack format
            if not port_utils.has_ip_address_defined(row.mac[0]):
                return False
            if event == self.ROW_UPDATE:
                return (old.chassis[0].name == self.agent.chassis and
                        not row.chassis)
            else:
                return row.chassis[0].name == self.agent.chassis
        except (IndexError, AttributeError):
            return False

    def _run(self, event, row, old):
        if row.type != constants.OVN_CHASSISREDIRECT_VIF_PORT_TYPE:
            return
        with _SYNC_STATE_LOCK.read_lock():
            self.agent.withdraw_ip(row, cr_lrp=True)


class SubnetRouterAttachedEvent(base_watcher.PortBindingChassisEvent):
    def __init__(self, bgp_agent):
        events = (self.ROW_UPDATE, self.ROW_CREATE,)
        super(SubnetRouterAttachedEvent, self).__init__(
            bgp_agent, events)

    def match_fn(self, event, row, old):
        try:
            if event == self.ROW_UPDATE:
                return (not row.chassis and
                        not row.logical_port.startswith('lrp-') and
                        row.external_ids[constants.OVN_EVPN_VNI_EXT_ID_KEY] and
                        row.external_ids[constants.OVN_EVPN_AS_EXT_ID_KEY] and
                        (not old.external_ids.get(
                            constants.OVN_EVPN_VNI_EXT_ID_KEY) or
                         not old.external_ids.get(
                             constants.constants.OVN_EVPN_AS_EXT_ID_KEY)))
            else:
                return (not row.chassis and
                        not row.logical_port.startswith('lrp-') and
                        row.external_ids[constants.OVN_EVPN_VNI_EXT_ID_KEY] and
                        row.external_ids[constants.OVN_EVPN_AS_EXT_ID_KEY])
        except (IndexError, AttributeError, KeyError):
            return False

    def _run(self, event, row, old):
        if row.type != constants.OVN_PATCH_VIF_PORT_TYPE:
            return
        with _SYNC_STATE_LOCK.read_lock():
            if row.nat_addresses:
                self.agent.expose_ip(row)
            else:
                self.agent.expose_subnet(row)


class SubnetRouterDetachedEvent(base_watcher.PortBindingChassisEvent):
    def __init__(self, bgp_agent):
        events = (self.ROW_UPDATE, self.ROW_DELETE,)
        super(SubnetRouterDetachedEvent, self).__init__(
            bgp_agent, events)

    def match_fn(self, event, row, old):
        try:
            if event == self.ROW_UPDATE:
                return (not row.chassis and
                        not row.logical_port.startswith('lrp-') and
                        old.external_ids[constants.OVN_EVPN_VNI_EXT_ID_KEY] and
                        old.external_ids[constants.OVN_EVPN_AS_EXT_ID_KEY] and
                        (not row.external_ids.get(
                            constants.OVN_EVPN_VNI_EXT_ID_KEY) or
                         not row.external_ids.get(
                             constants.OVN_EVPN_AS_EXT_ID_KEY)))
            else:
                return (not row.chassis and
                        not row.logical_port.startswith('lrp-') and
                        row.external_ids[constants.OVN_EVPN_VNI_EXT_ID_KEY] and
                        row.external_ids[constants.OVN_EVPN_AS_EXT_ID_KEY])
        except (IndexError, AttributeError, KeyError):
            return False

    def _run(self, event, row, old):
        if row.type != constants.OVN_PATCH_VIF_PORT_TYPE:
            return
        with _SYNC_STATE_LOCK.read_lock():
            if row.nat_addresses:
                self.agent.withdraw_ip(row)
            else:
                self.agent.withdraw_subnet(row)


class TenantPortCreatedEvent(base_watcher.PortBindingChassisEvent):
    def __init__(self, bgp_agent):
        events = (self.ROW_UPDATE,)
        super(TenantPortCreatedEvent, self).__init__(
            bgp_agent, events)

    def match_fn(self, event, row, old):
        try:
            # single and dual-stack format
            if not port_utils.has_ip_address_defined(row.mac[0]):
                return False
            return (not old.chassis and row.chassis and
                    self.agent.ovn_local_lrps != [])
        except (IndexError, AttributeError):
            return False

    def _run(self, event, row, old):
        if row.type not in (constants.OVN_VM_VIF_PORT_TYPE,
                            constants.OVN_VIRTUAL_VIF_PORT_TYPE):
            return
        with _SYNC_STATE_LOCK.read_lock():
            ips = row.mac[0].split(' ')[1:]
            self.agent.expose_remote_ip(ips, row)


class TenantPortDeletedEvent(base_watcher.PortBindingChassisEvent):
    def __init__(self, bgp_agent):
        events = (self.ROW_DELETE, self.ROW_UPDATE,)
        super(TenantPortDeletedEvent, self).__init__(
            bgp_agent, events)

    def match_fn(self, event, row, old):
        try:
            # single and dual-stack format
            if not port_utils.has_ip_address_defined(row.mac[0]):
                return False
            if event == self.ROW_UPDATE:
                return (old.chassis and not row.chassis and
                        self.agent.ovn_local_lrps != [])
            else:
                return (self.agent.ovn_local_lrps != [])
        except (IndexError, AttributeError):
            return False

    def _run(self, event, row, old):
        if row.type not in (constants.OVN_VM_VIF_PORT_TYPE,
                            constants.OVN_VIRTUAL_VIF_PORT_TYPE):
            return
        with _SYNC_STATE_LOCK.read_lock():
            ips = row.mac[0].split(' ')[1:]
            self.agent.withdraw_remote_ip(ips, row)


class LocalnetCreateDeleteEvent(base_watcher.PortBindingChassisEvent):
    def __init__(self, bgp_agent):
        events = (self.ROW_CREATE, self.ROW_DELETE,)
        super(LocalnetCreateDeleteEvent, self).__init__(
            bgp_agent, events)

    def match_fn(self, event, row, old):
        return row.type == constants.OVN_LOCALNET_VIF_PORT_TYPE

    def _run(self, event, row, old):
        with _SYNC_STATE_LOCK.read_lock():
            self.agent.sync()



# Copyright 2022 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from ovn_bgp_agent import constants
from ovn_bgp_agent.drivers.openstack.watchers import evpn_watcher
from ovn_bgp_agent.tests import base as test_base
from ovn_bgp_agent.tests import utils


class TestPortBindingChassisCreatedEvent(test_base.TestCase):

    def setUp(self):
        super(TestPortBindingChassisCreatedEvent, self).setUp()
        self.chassis = '935f91fa-b8f8-47b9-8b1b-3a7a90ef7c26'
        self.agent = mock.Mock(chassis=self.chassis)
        self.event = evpn_watcher.PortBindingChassisCreatedEvent(self.agent)

    def test_match_fn(self):
        ch = utils.create_row(name=self.chassis)
        row = utils.create_row(
            type=constants.OVN_CHASSISREDIRECT_VIF_PORT_TYPE,
            chassis=[ch], mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'])
        old = utils.create_row(chassis=[])
        self.assertTrue(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_not_single_or_dual_stack(self):
        ch = utils.create_row(name=self.chassis)
        row = utils.create_row(
            type=constants.OVN_CHASSISREDIRECT_VIF_PORT_TYPE,
            chassis=[ch], mac=['aa:bb:cc:dd:ee:ff'])
        old = utils.create_row(chassis=[])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_old_chassis_set(self):
        ch = utils.create_row(name=self.chassis)
        row = utils.create_row(
            type=constants.OVN_CHASSISREDIRECT_VIF_PORT_TYPE,
            chassis=[ch], mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'])
        old = utils.create_row(chassis=self.chassis)
        self.assertFalse(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_index_error(self):
        row = utils.create_row(
            type=constants.OVN_CHASSISREDIRECT_VIF_PORT_TYPE,
            chassis=[], mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'])
        old = utils.create_row(chassis=[])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_wrong_type(self):
        ch = utils.create_row(name=self.chassis)
        row = utils.create_row(type='farofa', chassis=[ch],
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'])
        old = utils.create_row(chassis=[])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, old))

    def test_run(self):
        row = utils.create_row(
            type=constants.OVN_CHASSISREDIRECT_VIF_PORT_TYPE)
        self.event.run(mock.Mock(), row, mock.Mock())
        self.agent.expose_ip.assert_called_once_with(row, cr_lrp=True)

    def test_run_dual_stack(self):
        row = utils.create_row(
            type=constants.OVN_CHASSISREDIRECT_VIF_PORT_TYPE)
        self.event.run(mock.Mock(), row, mock.Mock())
        self.agent.expose_ip.assert_called_once_with(row, cr_lrp=True)

    def test_run_wrong_type(self):
        row = utils.create_row(type='farofa')
        self.event.run(mock.Mock(), row, mock.Mock())
        self.agent.expose_ip.assert_not_called()


class TestPortBindingChassisDeletedEvent(test_base.TestCase):

    def setUp(self):
        super(TestPortBindingChassisDeletedEvent, self).setUp()
        self.chassis = '935f91fa-b8f8-47b9-8b1b-3a7a90ef7c26'
        self.agent = mock.Mock(chassis=self.chassis)
        self.event = evpn_watcher.PortBindingChassisDeletedEvent(self.agent)

    def test_match_fn(self):
        ch = utils.create_row(name=self.chassis)
        row = utils.create_row(
            type=constants.OVN_CHASSISREDIRECT_VIF_PORT_TYPE,
            chassis=[ch], mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'])
        old = utils.create_row(chassis=[])
        self.assertTrue(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_not_single_or_dual_stack(self):
        ch = utils.create_row(name=self.chassis)
        row = utils.create_row(
            type=constants.OVN_CHASSISREDIRECT_VIF_PORT_TYPE,
            chassis=[ch], mac=['aa:bb:cc:dd:ee:ff'])
        old = utils.create_row(chassis=[])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_update(self):
        event = self.event.ROW_UPDATE
        ch = utils.create_row(name=self.chassis)
        row = utils.create_row(
            type=constants.OVN_CHASSISREDIRECT_VIF_PORT_TYPE,
            chassis=[], mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'])
        old = utils.create_row(chassis=[ch])
        self.assertTrue(self.event.match_fn(event, row, old))

    def test_match_fn_update_old_chassis_set(self):
        event = self.event.ROW_UPDATE
        ch = utils.create_row(name=self.chassis)
        row = utils.create_row(
            type=constants.OVN_CHASSISREDIRECT_VIF_PORT_TYPE,
            chassis=[ch], mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'])
        old = utils.create_row(chassis=self.chassis)
        self.assertFalse(self.event.match_fn(event, row, old))

    def test_match_fn_index_error(self):
        row = utils.create_row(
            type=constants.OVN_CHASSISREDIRECT_VIF_PORT_TYPE,
            chassis=[], mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'])
        old = utils.create_row(chassis=[])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_wrong_type(self):
        row = utils.create_row(
            type='farofa',
            chassis=[], mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'])
        old = utils.create_row(chassis=[])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, old))

    def test_run(self):
        row = utils.create_row(
            type=constants.OVN_CHASSISREDIRECT_VIF_PORT_TYPE)
        self.event.run(mock.Mock(), row, mock.Mock())
        self.agent.withdraw_ip.assert_called_once_with(row, cr_lrp=True)

    def test_run_wrong_type(self):
        row = utils.create_row(type='farofa')
        self.event.run(mock.Mock(), row, mock.Mock())
        self.agent.withdraw_ip.assert_not_called()


class TestSubnetRouterAttachedEvent(test_base.TestCase):

    def setUp(self):
        super(TestSubnetRouterAttachedEvent, self).setUp()
        self.chassis = '935f91fa-b8f8-47b9-8b1b-3a7a90ef7c26'
        self.agent = mock.Mock(chassis=self.chassis)
        self.event = evpn_watcher.SubnetRouterAttachedEvent(self.agent)

    def test_match_fn(self):
        ext_ids = {constants.OVN_EVPN_VNI_EXT_ID_KEY: 'fake-vni-id',
                   constants.OVN_EVPN_AS_EXT_ID_KEY: 'fake-as-id'}
        row = utils.create_row(
            chassis=[], logical_port='fake-lrp', external_ids=ext_ids)
        self.assertTrue(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_chassis_set(self):
        ch = utils.create_row(name=self.chassis)
        ext_ids = {constants.OVN_EVPN_VNI_EXT_ID_KEY: 'fake-vni-id',
                   constants.OVN_EVPN_AS_EXT_ID_KEY: 'fake-as-id'}
        row = utils.create_row(
            chassis=[ch], logical_port='fake-lrp', external_ids=ext_ids)
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_lrp(self):
        ext_ids = {constants.OVN_EVPN_VNI_EXT_ID_KEY: 'fake-vni-id',
                   constants.OVN_EVPN_AS_EXT_ID_KEY: 'fake-as-id'}
        row = utils.create_row(
            chassis=[], logical_port='lrp-fake', external_ids=ext_ids)
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_missing_ext_ids(self):
        row = utils.create_row(
            chassis=[], logical_port='fake-lrp', external_ids={})
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_update(self):
        event = self.event.ROW_UPDATE
        ext_ids = {constants.OVN_EVPN_VNI_EXT_ID_KEY: 'fake-vni-id',
                   constants.OVN_EVPN_AS_EXT_ID_KEY: 'fake-as-id'}
        row = utils.create_row(
            chassis=[], logical_port='fake-lrp', external_ids=ext_ids)
        old = utils.create_row(external_ids={})
        self.assertTrue(self.event.match_fn(event, row, old))

    def test_match_fn_update_chassis_set(self):
        event = self.event.ROW_UPDATE
        ch = utils.create_row(name=self.chassis)
        ext_ids = {constants.OVN_EVPN_VNI_EXT_ID_KEY: 'fake-vni-id',
                   constants.OVN_EVPN_AS_EXT_ID_KEY: 'fake-as-id'}
        row = utils.create_row(
            chassis=[ch], logical_port='fake-lrp', external_ids=ext_ids)
        old = utils.create_row(external_ids={})
        self.assertFalse(self.event.match_fn(event, row, old))

    def test_match_fn_update_lrp(self):
        event = self.event.ROW_UPDATE
        ext_ids = {constants.OVN_EVPN_VNI_EXT_ID_KEY: 'fake-vni-id',
                   constants.OVN_EVPN_AS_EXT_ID_KEY: 'fake-as-id'}
        row = utils.create_row(
            chassis=[], logical_port='lrp-fake', external_ids=ext_ids)
        old = utils.create_row(external_ids={})
        self.assertFalse(self.event.match_fn(event, row, old))

    def test_match_fn_update_missing_ext_ids(self):
        event = self.event.ROW_UPDATE
        row = utils.create_row(
            chassis=[], logical_port='fake-lrp', external_ids={})
        old = utils.create_row(external_ids={})
        self.assertFalse(self.event.match_fn(event, row, old))

    def test_match_fn_update_old_ext_ids(self):
        event = self.event.ROW_UPDATE
        ext_ids = {constants.OVN_EVPN_VNI_EXT_ID_KEY: 'fake-vni-id',
                   constants.OVN_EVPN_AS_EXT_ID_KEY: 'fake-as-id'}
        row = utils.create_row(
            chassis=[], logical_port='fake-lrp', external_ids=ext_ids)
        old = utils.create_row(external_ids=ext_ids)
        self.assertFalse(self.event.match_fn(event, row, old))

    def test_run(self):
        row = utils.create_row(
            type=constants.OVN_PATCH_VIF_PORT_TYPE, nat_addresses=[])
        self.event.run(mock.Mock(), row, mock.Mock())
        self.agent.expose_subnet.assert_called_once_with(row)

    def test_run_nat_addresses(self):
        row = utils.create_row(
            type=constants.OVN_PATCH_VIF_PORT_TYPE,
            nat_addresses=['10.10.1.16'])
        self.event.run(mock.Mock(), row, mock.Mock())
        self.agent.expose_ip.assert_called_once_with(row)

    def test_run_wrong_type(self):
        row = utils.create_row(type='farofa')
        self.event.run(mock.Mock(), row, mock.Mock())
        self.agent.expose_ip.assert_not_called()
        self.agent.expose_subnet.assert_not_called()


class TestSubnetRouterDetachedEvent(test_base.TestCase):

    def setUp(self):
        super(TestSubnetRouterDetachedEvent, self).setUp()
        self.chassis = '935f91fa-b8f8-47b9-8b1b-3a7a90ef7c26'
        self.agent = mock.Mock(chassis=self.chassis)
        self.event = evpn_watcher.SubnetRouterDetachedEvent(self.agent)

    def test_match_fn(self):
        ext_ids = {constants.OVN_EVPN_VNI_EXT_ID_KEY: 'fake-vni-id',
                   constants.OVN_EVPN_AS_EXT_ID_KEY: 'fake-as-id'}
        row = utils.create_row(
            chassis=[], logical_port='fake-lrp', external_ids=ext_ids)
        self.assertTrue(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_chassis_set(self):
        ch = utils.create_row(name=self.chassis)
        ext_ids = {constants.OVN_EVPN_VNI_EXT_ID_KEY: 'fake-vni-id',
                   constants.OVN_EVPN_AS_EXT_ID_KEY: 'fake-as-id'}
        row = utils.create_row(
            chassis=[ch], logical_port='fake-lrp', external_ids=ext_ids)
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_lrp(self):
        ext_ids = {constants.OVN_EVPN_VNI_EXT_ID_KEY: 'fake-vni-id',
                   constants.OVN_EVPN_AS_EXT_ID_KEY: 'fake-as-id'}
        row = utils.create_row(
            chassis=[], logical_port='lrp-fake', external_ids=ext_ids)
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_missing_ext_ids(self):
        row = utils.create_row(
            chassis=[], logical_port='fake-lrp', external_ids={})
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_update(self):
        event = self.event.ROW_UPDATE
        ext_ids = {constants.OVN_EVPN_VNI_EXT_ID_KEY: 'fake-vni-id',
                   constants.OVN_EVPN_AS_EXT_ID_KEY: 'fake-as-id'}
        row = utils.create_row(
            chassis=[], logical_port='fake-lrp', external_ids={})
        old = utils.create_row(external_ids=ext_ids)
        self.assertTrue(self.event.match_fn(event, row, old))

    def test_match_fn_update_chassis_set(self):
        event = self.event.ROW_UPDATE
        ch = utils.create_row(name=self.chassis)
        ext_ids = {constants.OVN_EVPN_VNI_EXT_ID_KEY: 'fake-vni-id',
                   constants.OVN_EVPN_AS_EXT_ID_KEY: 'fake-as-id'}
        row = utils.create_row(
            chassis=[ch], logical_port='fake-lrp', external_ids={})
        old = utils.create_row(external_ids=ext_ids)
        self.assertFalse(self.event.match_fn(event, row, old))

    def test_match_fn_update_lrp(self):
        event = self.event.ROW_UPDATE
        ext_ids = {constants.OVN_EVPN_VNI_EXT_ID_KEY: 'fake-vni-id',
                   constants.OVN_EVPN_AS_EXT_ID_KEY: 'fake-as-id'}
        row = utils.create_row(
            chassis=[], logical_port='lrp-fake', external_ids={})
        old = utils.create_row(external_ids=ext_ids)
        self.assertFalse(self.event.match_fn(event, row, old))

    def test_match_fn_update_missing_ext_ids(self):
        event = self.event.ROW_UPDATE
        row = utils.create_row(
            chassis=[], logical_port='fake-lrp', external_ids={})
        old = utils.create_row(external_ids={})
        self.assertFalse(self.event.match_fn(event, row, old))

    def test_match_fn_update_row_ext_ids(self):
        event = self.event.ROW_UPDATE
        ext_ids = {constants.OVN_EVPN_VNI_EXT_ID_KEY: 'fake-vni-id',
                   constants.OVN_EVPN_AS_EXT_ID_KEY: 'fake-as-id'}
        row = utils.create_row(
            chassis=[], logical_port='fake-lrp', external_ids=ext_ids)
        old = utils.create_row(external_ids=ext_ids)
        self.assertFalse(self.event.match_fn(event, row, old))

    def test_run(self):
        row = utils.create_row(
            type=constants.OVN_PATCH_VIF_PORT_TYPE, nat_addresses=[])
        self.event.run(mock.Mock(), row, mock.Mock())
        self.agent.withdraw_subnet.assert_called_once_with(row)

    def test_run_nat_addresses(self):
        row = utils.create_row(
            type=constants.OVN_PATCH_VIF_PORT_TYPE,
            nat_addresses=['10.10.1.16'])
        self.event.run(mock.Mock(), row, mock.Mock())
        self.agent.withdraw_ip.assert_called_once_with(row)

    def test_run_wrong_type(self):
        row = utils.create_row(type='farofa')
        self.event.run(mock.Mock(), row, mock.Mock())
        self.agent.withdraw_ip.assert_not_called()
        self.agent.withdraw_subnet.assert_not_called()


class TestTenantPortCreatedEvent(test_base.TestCase):

    def setUp(self):
        super(TestTenantPortCreatedEvent, self).setUp()
        self.chassis = '935f91fa-b8f8-47b9-8b1b-3a7a90ef7c26'
        self.agent = mock.Mock(chassis=self.chassis)
        self.agent.ovn_local_lrps = ['172.24.100.111']
        self.event = evpn_watcher.TenantPortCreatedEvent(self.agent)

    def test_match_fn(self):
        ch = utils.create_row(name=self.chassis)
        row = utils.create_row(mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'],
                               chassis=[ch])
        old = utils.create_row(chassis=[])
        self.assertTrue(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_not_single_or_dual_stack(self):
        ch = utils.create_row(name=self.chassis)
        row = utils.create_row(mac=['aa:bb:cc:dd:ee:ff'],
                               chassis=[ch])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_old_chassis_set(self):
        ch = utils.create_row(name=self.chassis)
        row = utils.create_row(mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'],
                               chassis=[ch])
        old = utils.create_row(chassis=[ch])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_empty_ovn_local_lrps(self):
        ch = utils.create_row(name=self.chassis)
        self.agent.ovn_local_lrps = []
        row = utils.create_row(mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'],
                               chassis=[ch])
        old = utils.create_row(chassis=[])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_index_error(self):
        row = utils.create_row(mac=[])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_run(self):
        row = utils.create_row(type=constants.OVN_VM_VIF_PORT_TYPE,
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'])
        self.event.run(mock.Mock(), row, mock.Mock())
        self.agent.expose_remote_ip.assert_called_once_with(
            ['10.10.1.16'], row)

    def test_run_dual_stack(self):
        row = utils.create_row(
            type=constants.OVN_VM_VIF_PORT_TYPE,
            mac=['aa:bb:cc:dd:ee:ff 10.10.1.16 2002::1234:abcd:ffff:c0a8:101'])
        self.event.run(mock.Mock(), row, mock.Mock())
        self.agent.expose_remote_ip.assert_called_once_with(
            ['10.10.1.16', '2002::1234:abcd:ffff:c0a8:101'], row)

    def test_run_wrong_type(self):
        row = utils.create_row(type='brigadeiro')
        self.event.run(mock.Mock(), row, mock.Mock())
        self.agent.expose_remote_ip.assert_not_called()


class TestTenantPortDeletedEvent(test_base.TestCase):

    def setUp(self):
        super(TestTenantPortDeletedEvent, self).setUp()
        self.chassis = '935f91fa-b8f8-47b9-8b1b-3a7a90ef7c26'
        self.agent = mock.Mock(chassis=self.chassis)
        self.agent.ovn_local_lrps = ['172.24.100.111']
        self.event = evpn_watcher.TenantPortDeletedEvent(self.agent)

    def test_match_fn(self):
        ch = utils.create_row(name=self.chassis)
        row = utils.create_row(mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'],
                               chassis=[ch])
        self.assertTrue(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_not_single_or_dual_stack(self):
        ch = utils.create_row(name=self.chassis)
        row = utils.create_row(mac=['aa:bb:cc:dd:ee:ff'],
                               chassis=[ch])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_empty_ovn_local_lrps(self):
        ch = utils.create_row(name=self.chassis)
        self.agent.ovn_local_lrps = []
        row = utils.create_row(mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'],
                               chassis=[ch])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_update_old_chassis_set(self):
        event = self.event.ROW_UPDATE
        ch = utils.create_row(name=self.chassis)
        row = utils.create_row(mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'],
                               chassis=[ch])
        old = utils.create_row(chassis=[ch])
        self.assertFalse(self.event.match_fn(event, row, old))

    def test_match_fn_update_empty_ovn_local_lrps(self):
        event = self.event.ROW_UPDATE
        ch = utils.create_row(name=self.chassis)
        self.agent.ovn_local_lrps = []
        row = utils.create_row(mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'],
                               chassis=[ch])
        old = utils.create_row(chassis=[])
        self.assertFalse(self.event.match_fn(event, row, old))

    def test_match_fn_index_error(self):
        row = utils.create_row(mac=[])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_run(self):
        row = utils.create_row(type=constants.OVN_VM_VIF_PORT_TYPE,
                               mac=['aa:bb:cc:dd:ee:ff 10.10.1.16'])
        self.event.run(mock.Mock(), row, mock.Mock())
        self.agent.withdraw_remote_ip.assert_called_once_with(
            ['10.10.1.16'], row)

    def test_run_dual_stack(self):
        row = utils.create_row(
            type=constants.OVN_VM_VIF_PORT_TYPE,
            mac=['aa:bb:cc:dd:ee:ff 10.10.1.16 2002::1234:abcd:ffff:c0a8:101'])
        self.event.run(mock.Mock(), row, mock.Mock())
        self.agent.withdraw_remote_ip.assert_called_once_with(
            ['10.10.1.16', '2002::1234:abcd:ffff:c0a8:101'], row)

    def test_run_wrong_type(self):
        row = utils.create_row(type='brigadeiro')
        self.event.run(mock.Mock(), row, mock.Mock())
        self.agent.withdraw_remote_ip.assert_not_called()


class TestLocalnetCreateDeleteEvent(test_base.TestCase):

    def setUp(self):
        super(TestLocalnetCreateDeleteEvent, self).setUp()
        self.agent = mock.Mock()
        self.event = evpn_watcher.LocalnetCreateDeleteEvent(self.agent)

    def test_match_fn(self):
        row = utils.create_row(type=constants.OVN_LOCALNET_VIF_PORT_TYPE)
        self.assertTrue(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_not_match(self):
        row = utils.create_row(type=constants.OVN_VM_VIF_PORT_TYPE)
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_run(self):
        self.event.run(mock.Mock(), mock.Mock(), mock.Mock())
        self.agent.sync.assert_called_once()



# Copyright 2022 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from ovn_bgp_agent.drivers.openstack.watchers import base_watcher
from ovn_bgp_agent.tests import base as test_base


class TestChassisCreateEvent(test_base.TestCase):
    _event = base_watcher.ChassisCreateEvent

    def setUp(self):
        super(TestChassisCreateEvent, self).setUp()
        self.chassis = '935f91fa-b8f8-47b9-8b1b-3a7a90ef7c26'
        self.agent = mock.Mock(chassis=self.chassis)
        self.event = self._event(self.agent)

    def test_run(self):
        self.assertTrue(self.event.first_time)
        self.event.run(mock.Mock(), mock.Mock(), mock.Mock())

        self.assertFalse(self.event.first_time)
        self.agent.sync.assert_not_called()

    def test_run_not_first_time(self):
        self.event.first_time = False
        self.event.run(mock.Mock(), mock.Mock(), mock.Mock())
        self.agent.sync.assert_called_once_with()


class TestChassisPrivateCreateEvent(TestChassisCreateEvent):
    _event = base_watcher.ChassisPrivateCreateEvent



# Copyright 2024 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def get_from_external_ids(row, key):
    try:
        return row.external_ids[key]
    except (AttributeError, KeyError):
        pass


def ip_matches_in_row(row, ip, key):
    """Return True if given ip is in external_ids under given key.

    Return also True if passed ip is None and key is not present.

    Return None if external_ids is not present in row.

    Otherwise return False
    """
    try:
        return ip == row.external_ids.get(key)
    except AttributeError:
        pass



# Copyright 2024 team.blue/nl
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import netaddr

from oslo_config import cfg
from oslo_log import log as logging

from ovn_bgp_agent import constants
from ovn_bgp_agent.drivers.openstack.utils import driver_utils
from ovn_bgp_agent.drivers.openstack.utils import frr
from ovn_bgp_agent.drivers.openstack.utils import ovs
from ovn_bgp_agent import exceptions
from ovn_bgp_agent.utils import linux_net

CONF = cfg.CONF
LOG = logging.getLogger(__name__)

# dictionary to hold all evpn bridge classes
local_bridges: 'dict[str, EvpnBridge]' = {}

# dictionary to hold all vlandev mappings, based on port uuid
local_vlandevs: 'dict[str, VlanDev]' = {}


class EvpnBridge:
    def __init__(self, ovs_bridge: str, vni: int, evpn_opts: dict,
                 mode=constants.OVN_EVPN_TYPE_L3, ovs_flows: dict = None):

        if not CONF.evpn_local_ip:
            LOG.error("EVPN device must have an IP associated for the "
                      "VXLAN local ip")
            raise exceptions.ConfOptionRequired(option='evpn_local_ip')

        self.ovs_bridge = ovs_bridge
        self.vni = vni
        self.mode = mode
        self.ovs_flows = ovs_flows or {}

        self.vrf_name = '%s%s' % (constants.OVN_EVPN_VRF_PREFIX, vni)
        self.bridge_name = '%s%s' % (constants.OVN_EVPN_BRIDGE_PREFIX, vni)
        self.vxlan_name = '%s%s' % (constants.OVN_EVPN_VXLAN_PREFIX, vni)
        self.local_ip = CONF.evpn_local_ip
        self.evpn_opts = dict(
            vni=self.vni,
            local_ip=self.local_ip,
            vrf_name=self.vrf_name,
            redistribute=['connected', 'kernel'],
        )
        self.evpn_opts.update(evpn_opts)

        self.vlans: dict[str, 'VlanDev'] = {}
        self._setup_done = False

    def setup(self):
        if self._setup_done:
            return

        LOG.debug('Creating bridge %s', self.bridge_name)
        linux_net.ensure_bridge(self.bridge_name)

        LOG.debug('Creating vxlan interface %s for vni %s with local ip %s',
                  self.vxlan_name, self.vni, self.local_ip)
        linux_net.ensure_vxlan(self.vxlan_name, self.vni, self.local_ip,
                               CONF.evpn_udp_dstport)

        LOG.debug('Connect vxlan interface %s to bridge %s',
                  self.vxlan_name, self.bridge_name)
        linux_net.set_master_for_device(self.vxlan_name, self.bridge_name)

        LOG.debug('Disable learning for vxlan device %s', self.vxlan_name)
        linux_net.disable_learning_vxlan_intf(self.vxlan_name)

        LOG.debug('Configure FRR VRF (add)')
        frr.vrf_reconfigure(self.evpn_opts, 'add-vrf')

        if self.mode == constants.OVN_EVPN_TYPE_L3:
            LOG.debug('Create L3 EVPN devices')
            linux_net.ensure_vrf(self.vrf_name, self.vni)

            LOG.debug('Attach bridge %s to vrf %s',
                      self.bridge_name, self.vrf_name)
            linux_net.set_master_for_device(self.bridge_name, self.vrf_name)

        self._setup_done = True

    def _eval_disconnect(self):
        if not self._setup_done:
            return

        connected = [1 for v in self.vlans.values() if v._setup_done]
        if len(connected) == 0:
            # All vlan interfaces are unprovisioned, proceed with disconnect
            return self.disconnect()

        LOG.debug('No disconnect needed, there are still %s vlans active',
                  len(connected))

    def disconnect(self):
        LOG.info('Disconnecting evpn bridge %s',
                 self.vrf_name)

        disconnect_devices = [self.bridge_name, self.vxlan_name]
        if CONF.delete_vrf_on_disconnect:
            disconnect_devices.append(self.vrf_name)

        for devname in disconnect_devices:
            LOG.info('Delete device %s', devname)
            linux_net.delete_device(devname)

        if CONF.delete_vrf_on_disconnect:
            # We need to do the frr reconfigure after deleting all devices.
            # otherwise, frr will throw an error that it can only delete
            # inactive vrf's
            LOG.debug('Configure FRR VRF (del)')
            frr.vrf_reconfigure(self.evpn_opts, action="del-vrf")

        self._setup_done = False

    def connect_vlan(self, port):
        vlan_tag = driver_utils.get_port_vlan(port)
        if vlan_tag not in self.vlans:
            self.vlans[vlan_tag] = VlanDev(self, port)

        return self.vlans[vlan_tag]

    def get_vlan(self, vlan: 'int|str|None') -> 'VlanDev':
        if vlan is None:
            vlan = constants.VLAN_ID_UNTAGGED

        return self.vlans[str(vlan)]


class VlanDev:
    def __init__(self, bridge: 'EvpnBridge', port):
        self.bridge = bridge
        self.port = port  # localnet port

        uuid = str(port.uuid)[0:11]
        self.veth_vrf = constants.OVN_EVPN_VETH_VRF_UUID_PREFIX + uuid
        self.veth_ovs = constants.OVN_EVPN_VETH_OVS_UUID_PREFIX + uuid

        if uuid in local_vlandevs:
            # It already exists, but probably in another vni
            local_vlandevs[uuid].teardown()

        local_vlandevs[uuid] = self

        self.vlan_tag = driver_utils.get_port_vlan(port)

        # Will be filled by the @property with the mac address of
        # the vrf-vlan interface
        self._lladdr = None

        # boolean to indicate if setup is required for this interface
        self._setup_done = False
        self._veth_created = False

        # list of custom addresses to use during setup, before adding the
        # 169.254.x.x address, making another possibly public ip the
        # primary ip in traceroutes
        self._custom_ips = set()

        # list with tuple of (task, args, kwargs) we should execute once the
        # setup has completed. For example to run frr config for
        # ipv6 neighbor discovery, or to add ip's to the interface.
        self._post_setup_tasks = []

        self._agent_routing_tables_routes = collections.defaultdict(list)
        self._route_table_routes = {}

    def _set_agent_cache(self, routing_tables_routes):
        if routing_tables_routes is not None:
            self._agent_routing_tables_routes = (
                routing_tables_routes[self.veth_vrf])

    @property
    def lladdr(self):
        if not self._lladdr:
            if not self._veth_created:
                self.setup()
            self._lladdr = linux_net.get_interface_address(self.veth_vrf)
        return self._lladdr

    def setup(self):
        if self._setup_done:
            return

        # Run the setup of the bridge.
        self.bridge.setup()

        LOG.debug('Create VLAN veth interface %s <-> %s',
                  self.veth_vrf, self.veth_ovs)
        linux_net.ensure_veth(self.veth_vrf, self.veth_ovs)
        self._veth_created = True

        # Connect the veth_ovs to ovs
        ovs_vlan_tag = self.vlan_tag
        if self.vlan_tag == '0':
            ovs_vlan_tag = None

        ovs.add_device_to_ovs_bridge(self.veth_ovs,
                                     self.bridge.ovs_bridge,
                                     vlan_tag=ovs_vlan_tag)

        # Connect veth to bridge for L2
        if self.bridge.mode == constants.OVN_EVPN_TYPE_L2:
            linux_net.set_master_for_device(self.veth_vrf,
                                            self.bridge.bridge_name)
            self._setup_done = True
            return

        # Connect veth to vrf for L3
        # Create vrf interface, connect bridge and veth_vrf to it.
        LOG.debug('Configure L3 for EVPN devices')
        linux_net.set_master_for_device(self.veth_vrf, self.bridge.vrf_name)

        if self._custom_ips:
            linux_net.add_ips_to_dev(self.veth_vrf, ips=list(self._custom_ips))

        # Add 169.254.x.x address to veth_vrf for ipv4 and ipv6
        linux_net.ensure_arp_ndp_enabled_for_bridge(
            self.veth_vrf, offset=int(self.vlan_tag), vlan_tag=self.vlan_tag
        )

        # Configure mac on the veth interface to be the same on all hosts
        offset = _offset_for_vni_and_vlan(self.bridge.vni, self.vlan_tag)
        linux_net.ensure_anycast_mac_for_interface(
            self.veth_vrf, offset=offset
        )

        # Make sure ipv4 and ipv6 forwarding is enabled
        linux_net.enable_routing_for_interfaces(self.veth_vrf,
                                                self.bridge.bridge_name)

        # As long as we use 169.254.x.x addresses, we require proxy arp to be
        # there for initial router discovery
        linux_net.enable_proxy_arp(self.veth_vrf)
        linux_net.enable_proxy_ndp(self.veth_vrf)

        ovs_ok = self._setup_ovs()
        if ovs_ok is False:
            LOG.error('Unable to setup ovs, a retry will pick it up.')
            return

        # Any post-setup tasks to run.
        for method, a, kw in self._post_setup_tasks:
            method(*a, **kw)

        self._setup_done = True

    def _setup_ovs(self):
        try:
            in_port = ovs.get_ovs_patch_port_ofport(self.port.name)
            LOG.debug('ovs in-port: %s', in_port)
        except Exception:
            return False

        ovs_flows = self.bridge.ovs_flows
        ovs_bridge = self.bridge.ovs_bridge

        pmm = ovs_flows[ovs_bridge].setdefault('port-mac-mapping', {})
        pmm[in_port] = self.lladdr

        ovs.ensure_mac_tweak_flows(ovs_bridge,
                                   self.lladdr,
                                   [in_port],
                                   constants.OVS_RULE_COOKIE)

        ovs.remove_extra_ovs_flows(ovs_flows, ovs_bridge,
                                   constants.OVS_RULE_COOKIE)

    def _eval_disconnect(self):
        if not self._setup_done:
            return

        if len(self._agent_routing_tables_routes) == 0:
            return self.disconnect()

        LOG.debug('No disconnect needed, there are still %s announcements',
                  len(self._agent_routing_tables_routes))

    def disconnect(self):
        LOG.info('Disconnecting vlan interface %s.%s',
                 self.bridge.vrf_name, self.vlan_tag)

        LOG.info('Remove device %s from ovs bridge %s', self.veth_ovs,
                 self.bridge.ovs_bridge)
        ovs.del_device_from_ovs_bridge(self.veth_ovs,
                                       self.bridge.ovs_bridge)

        LOG.info('Delete device %s', self.veth_vrf)
        linux_net.delete_device(self.veth_vrf)

        self._veth_created = False
        self._setup_done = False
        self.bridge._eval_disconnect()

    def teardown(self):
        LOG.info('Running teardown for vlandev %s (vni change)', self.veth_vrf)
        self.disconnect()
        del self.bridge.vlans[self.vlan_tag]

    def _run(self, method, *a, **kw):
        # Run the method if setup is done, otherwise, run them when setup
        # is called
        if not self._setup_done:
            self._post_setup_tasks.append([method, a, kw])
        else:
            method(*a, **kw)

    def process_dhcp_opts(self, dhcp_opts):
        '''Add IP's or router advertisements from configured dhcp options

        For networks that have DHCP enabled (and have lsp with dhcp options),
        we can add the IP of the gateway on our vlan interface. Then OVN is
        able to discover the IP.

        Also if IPv6 is configured, we should enable router advertisements
        through FRR (since it has it built-in anyway), so vm's can then
        receive the router information from the 'provider' side.
        '''

        for opt in dhcp_opts:
            ver = netaddr.IPNetwork(opt.cidr).version

            if opt.options.get('router', False):
                LOG.debug('Adding IPv%s gateway ip: %s',
                          ver, opt.options['router'])
                self.add_ips([opt.options['router']])

            if ver == 6 and opt.cidr:
                LOG.debug('Configure ipv6nd for %s and opts %s',
                          opt.cidr, opt.options)
                self.configure_nd(opt.cidr, opts=opt.options)

    def configure_nd(self, cidr, opts):
        self._run(frr.nd_reconfigure, self.veth_vrf, cidr, opts)

    def add_ips(self, ips: list):
        self._custom_ips.update(ips)
        self._run(linux_net.add_ips_to_dev, self.veth_vrf, ips=ips)

    def add_route(self, routing_tables_routes: 'dict | None', ip: str,
                  mac: 'str | None', via: 'str | None' = None):
        '''Will add route to the routing table for this vlan_dev

        Please make sure pass along the routing_tables_routes dictionary at
        least the first time a route is added (for example when exposing the
        lrp or lsp). Then with a expose_remote_ip we can re-use the reference
        from the agent set earlier.
        '''
        self.setup()  # setup the bridge and vlan, if not already done.
        self._set_agent_cache(routing_tables_routes)

        if self.bridge.mode != constants.OVN_EVPN_TYPE_L3:
            return

        mask = None
        if '/' in ip:
            ip, mask = ip.split('/')

        self._agent_routing_tables_routes.append({
            'ip': ip, 'mask': mask, 'mac': mac, 'via': via,
        })
        LOG.debug('Add route %s/%s via %s dev %s table %s',
                  ip, mask, via, self.veth_vrf, self.bridge.vni)
        linux_net.add_ip_route(self._route_table_routes, ip, self.bridge.vni,
                               self.veth_vrf, mask=mask, via=via)

        # When a floating ip is passed along, it is a set of mac
        # addresses, so ensure we are always processing a list.
        for lladdr in _ensure_list(mac):
            LOG.debug('Add neigh %s -> %s dev %s', ip, mac, self.veth_vrf)
            linux_net.add_ip_nei(ip, lladdr, self.veth_vrf)

    def del_route(self, routing_tables_routes: 'dict | None', ip: str,
                  lladdr: 'str | None' = None):
        '''Will remove the route from the routing table for this vlan_dev

        Please make sure pass along the routing_tables_routes dictionary at
        least the first time a route is added (for example when exposing the
        lrp or lsp). Then with a withdraw_remote_ip we can re-use the reference
        from the agent set earlier.

        lladdr is optional, as it will be fetched from the internal
        route table dictionary
        '''

        if self.bridge.mode != constants.OVN_EVPN_TYPE_L3:
            return
        self._set_agent_cache(routing_tables_routes)

        # When a floating ip is passed along, it is a set of mac
        # addresses, so ensure we are always processing a list.

        mask = None
        if '/' in ip:
            ip, mask = ip.split('/')

        route = _find_route_info(self._agent_routing_tables_routes, ip)

        # Remove route from vrf
        linux_net.del_ip_route(self._route_table_routes, ip, self.bridge.vni,
                               self.veth_vrf, mask=mask or route['mask'],
                               via=route['via'])

        # Remove any neighbor information for route.
        for mac in _ensure_list(lladdr or route['mac']):
            linux_net.del_ip_nei(ip, mac, self.veth_vrf)

        if route in self._agent_routing_tables_routes:
            self._agent_routing_tables_routes.remove(route)

        self._eval_disconnect()

    def cleanup_excessive_routes(self, routing_tables_routes: dict):
        if not self._setup_done:
            return

        self._set_agent_cache(routing_tables_routes)

        # Get all routes on host for our vrf and our veth_vrf
        intf_idx = linux_net.get_interface_index(self.veth_vrf)
        current_routes = dict([
            (r.get_attr('RTA_DST'), r)
            for r in linux_net._get_table_routes(self.bridge.vni)
            if r.get_attr('RTA_OIF') == intf_idx and
            r['type'] == constants.ROUTE_TYPE_UNICAST and
            r.get_attr('RTA_DST') not in ('fe80::') and
            not r.get_attr('RTA_DST').startswith(
                constants.NDP_IPV6_PREFIX)
        ])

        # Create set with prefixes currently on host
        prefixes = {r.get_attr('RTA_DST') for r in current_routes.values()}

        # Create set with prefixes we maintain
        exposed_prefixes = {r['ip']
                            for r in self._agent_routing_tables_routes}

        if len(prefixes - exposed_prefixes) == 0:
            LOG.debug('No excessive routes to remove.')

        for ip in prefixes - exposed_prefixes:
            LOG.info('Remove excessive route %s', ip)
            kernel_route = current_routes[ip]
            route = _find_route_info(self._agent_routing_tables_routes, ip)
            if ((route['mask'] and
                    int(route['mask']) != kernel_route['dst_len']) or
                    route['via'] != kernel_route.get_attr('RTA_GATEWAY')):
                self._agent_routing_tables_routes.append({
                    'ip': ip, 'mask': kernel_route['dst_len'], 'mac': None,
                    'via': kernel_route.get_attr('RTA_GATEWAY'),
                })

            self.del_route(routing_tables_routes, ip)


def _ensure_list(var):
    if var is None:
        return []

    if not isinstance(var, (list, tuple, set)):
        var = [var]

    return var


def _find_route_info(routes: 'list[dict]', ip: str):
    for r in routes:
        if r['ip'] == ip:
            return r

    return {'ip': ip, 'mask': None, 'mac': None, 'via': None}


def _offset_for_vni_and_vlan(vni: int, vlan: str):
    '''Generate a offset (in numeric system), based on the vni and vlan

    vni has range 1-16777214 (6 bytes)
    vlan has range 0-4094 (3 bytes)

    It will transform the vni to a 6 digit hex, append the 3 digit vlan hex
    and transform it back to a integer.
    '''
    if vni > 16777214:
        LOG.warning('Configured vni value %d is too big (range 1-16777214)',
                    vni)
        vni = vni % 0xffffff  # reset vni, to prevent overflow.

    if int(vlan) > 4094:
        LOG.warning('Configured vlan value %d is too big (range 0-4094)',
                    int(vlan))
        vlan = int(vlan) % 0xfff  # reset vlan, to prevent overflow.

    return int(''.join([
        ('%x' % vni).zfill(6),
        ('%x' % int(vlan)).zfill(4),
    ]), 16)


def setup(ovs_bridge, vni, evpn_opts, mode=constants.OVN_EVPN_TYPE_L3,
          ovs_flows={}) -> EvpnBridge:
    # This method will either create the EvpnBridge or return the one that
    # already exists for the current vni.

    vni = int(vni)  # make sure the vni is a int, for lookup purposes

    if local_bridges.get(vni, None) is None:
        local_bridges[vni] = EvpnBridge(ovs_bridge, vni, evpn_opts,
                                        mode=mode, ovs_flows=ovs_flows)
    else:
        local_bridges[vni].ovs_flows = ovs_flows

    return local_bridges[vni]


def lookup(ovs_bridge: str, vlan: str) -> EvpnBridge:
    if vlan is None:
        vlan = constants.VLAN_ID_UNTAGGED

    for br in local_bridges.values():
        if br.ovs_bridge == ovs_bridge:
            if str(vlan) in br.vlans:
                return br

    raise KeyError('Could not locate EVPN for bridge %s and/or vlan %s' % (
                   ovs_bridge, vlan))


def lookup_vlan(ovs_bridge: str, vlan: str) -> VlanDev:
    bridge = lookup(ovs_bridge, vlan)
    return bridge.get_vlan(vlan)
